import os, sys
from shutil import copyfile
from io import StringIO
from copy import deepcopy
import multiprocessing as mp

from Organisms.GA.Collection import Collection

class Population(Collection):
	"""
	This class stores all the clusters that are in the population.

	:param name: Name of the Collection. This can be any name you like, it is just to note what the collection is. All collections should have a unique name if possible to prevent confusion, however this will not break the program.
	:type  name: str.
	:param size: This is the number of clusters that should be in the collection. In this version of the Organisms program, the number of clusters in the population and made during Creation of Offspring should be consistant throughout the genetic algorithm process.
	:type  size: int
	:param user_initialised_population_folder: Indicates the directory to obtain clusters to place in the initial population. If None, there are no user created clusters to obtained. Default: None.
	:type  user_initialised_population_folder: str. 
	:param write_data: Write the clusters to disk in a database. Default: True.
	:type  write_data: bool.

	"""
	def __init__(self,name,size,user_initialised_population_folder=None,write_data=True):
		self.write_data = write_data
		have_database = write_collection_history = self.write_data
		super().__init__(name,size,have_database=have_database,path=None,write_collection_history=write_collection_history)
		self.user_initialised_population_folder = user_initialised_population_folder
		#############################
		# set up current_gen_details
		if self.write_data:
			self.current_population_details = 'current_population_details.txt'
			self.current_population_details_path = self.path+'/'+self.current_population_details

	def get_pool_folder_size(self,folder_to_look_at=False):
		"""
		This definition will count the number of clusters in the population that the user places into the
		GA before it runs. It does this by counting the number of numbered folders (these in this program
		hold information about each cluster created during a GA Run).
		This method is only needed before the GA begins.

		Inputs:
			folder_to_look_at (False/None/str.): Indicate the number of clusters in the population as stored on the disk.
		
		returns: len(clusters_in_population): The size of the population, measured by counting the number of folders named by a number (which this program will interpret as a cluster in the population).
		rtypes: int

		"""
		# obtain the folders that are named by a integer number. A numbered cluster indicates a cluster in the
		# population or if the offspring.
		if folder_to_look_at == False:
			folder_to_look_at = self.name
		clusters_in_population = []
		if folder_to_look_at == None or (not (os.path.exists(folder_to_look_at) and os.path.isdir(folder_to_look_at))):
			return 0
		for cluster in os.listdir(folder_to_look_at):
			if (os.path.isdir(os.path.join(folder_to_look_at, cluster)) and cluster.isdigit()): # If the cluster folder exists and the folder is numbered properl, then it is a cluster to be added into the GA program
				clusters_in_population.append(int(cluster))
		clusters_in_population.sort()
		# We will make sure that the clusters read from the cluster folders in the run folder are sequently
		# numbered, otherwise the user has done something wrong.
		noCluInPool = 1
		for cluster_name in clusters_in_population:
			if not noCluInPool == cluster_name:
				exit("Error: The clusters in the population are not sequently numbered. Please renumber your original clusters so they are in sequential order.")
			noCluInPool += 1
		# if there were no faults (i.e. the folders are sequently numbered), return the number.
		return len(clusters_in_population)

	##########################################################################
	############################# BACKUP METHODS #############################
	##########################################################################

	def backup_files(self):
		"""
		Backup the database and the current state file
		"""
		self.backup_database()
		self.backup_state_file()

	def remove_backup_files(self):
		"""
		Remove the backup the database and the current state file
		"""
		self.remove_backup_database()
		self.remove_backup_state_file()

	##########################################################################
	####################### current_state_file METHODS #######################
	##########################################################################

	def current_state_file(self,generation_number):
		"""
		Write the current state file for the population for this generation.

		:param generation_number: The curent generation.
		:type  generation_number: int

		"""
		if not self.have_database:
			return
		with open(self.path+'/'+self.current_population_details,'w') as Current_Population_details_TXT:
			Current_Population_details_TXT.write("GA Iteration: " + str(generation_number) + "\n")
			Current_Population_details_TXT.write("Clusters in Pool:\t")
			for cluster in self:
				Current_Population_details_TXT.write(str(cluster.name)+"\t")
			Current_Population_details_TXT.write("\n")
			Current_Population_details_TXT.write("Energies of Clusters:\t")
			for cluster in self:
				Current_Population_details_TXT.write(str(cluster.energy) + "\t")
			Current_Population_details_TXT.write("\n")

	def get_data_from_current_state_file(self,current_state_file):
		"""
		Get data about the current population from the state file on disk.

		:param current_state_file: The path to the curent state file.
		:type  current_state_file: str.

		:returns the generation the state file was made, a list of the names of the clusters in the population, and a list of the energies of the clusters in the population.
		:rtype   float, list of int, and list of float
		"""
		with open(current_state_file,'r') as current_state_fileTXT:
			generation_data = current_state_fileTXT.readline()
			generation = int(generation_data.rstrip().split(':')[1])
			cluster_names_data = current_state_fileTXT.readline()
			cluster_names = [int(name) for name in cluster_names_data.rstrip().split(':')[1].split()]
			cluster_energies_data = current_state_fileTXT.readline()
			cluster_energies = [float(energy) for energy in cluster_energies_data.rstrip().split(':')[1].split()]
		return generation, cluster_names, cluster_energies

	def is_data_in_current_state_file(self,current_state_file):
		try:
			self.get_data_from_current_state_file(current_state_file)
			return True
		except Exception as exception:
			print('------------------------------------------------------------------------')
			print('Could not get all the required data from the population state file (from '+str(current_state_file)+').')
			print('The following message is the exception message.')
			print()
			print(exception)
			print()
			print('The GA have some tricks up its sleeve to repair this issue.')
			print('------------------------------------------------------------------------')
			return False

	def backup_state_file(self):
		"""
		This method will make a backup of all the clusters in the Collection as a ASE database
		"""
		if self.have_database:
			copyfile(self.path+'/'+self.current_population_details, self.path+'/'+self.current_population_details+'.backup')
			os.remove(self.path+'/'+self.current_population_details)

	def remove_backup_state_file(self):
		"""
		This method will remove the backup of all the clusters in the Collection, which will be in the format of a ASE database.
		"""
		if self.have_database:
			os.remove(self.path+'/'+self.current_population_details+'.backup')

	def remove_backup_state_file_if_exists(self):
		"""
		This method will remove the backup of all the clusters in the Collection, which will be in the format of a ASE database.
		"""
		if self.have_database:
			if os.path.exists(self.path+'/'+self.current_population_details+'.backup'):
				os.remove(self.path+'/'+self.current_population_details+'.backup')

	def move_backup_to_current_files(self):
		"""
		This method reove the last state file if one exists, and replace it with the backup.
		"""
		if not self.have_database:
			return
		if os.path.exists(self.path+'/'+self.current_population_details):
			os.remove(self.path+'/'+self.current_population_details)
		copyfile(self.path+'/'+self.current_population_details+'.backup', self.path+'/'+self.current_population_details)
		os.remove(self.path+'/'+self.current_population_details+'.backup')

	def repair_current_state_file(self,generation_number,cluster_names,cluster_energies):
		"""
		Write the current state file for the population for this generation. 
		This method is used if the state file or a back up could not be found or were incomplete.
		This method will get the data from the information from the energyprofile.txt file

		:param generation_number: The current generation.
		:type  generation_number: int
		:param cluster_names: The names of the clusters in the population.
		:type  cluster_names: list of ints.
		:param cluster_energies: The energies of the clusters in the population.
		:type  cluster_energies: list of floats

		"""
		if not self.have_database:
			return
		path_to_current_population_details = self.path+'/'+self.current_population_details
		if os.path.exists(path_to_current_population_details):
			copyfile(path_to_current_population_details,path_to_current_population_details+'.original_before_restart.txt')
			os.remove(path_to_current_population_details)
		with open(path_to_current_population_details,'w') as Current_Population_details_TXT:
			Current_Population_details_TXT.write("GA Iteration: " + str(generation_number) + "\n")
			Current_Population_details_TXT.write("Clusters in Pool:\t")
			for cluster_name in cluster_names:
				Current_Population_details_TXT.write(str(cluster_name)+"\t")
			Current_Population_details_TXT.write("\n")
			Current_Population_details_TXT.write("Energies of Clusters:\t")
			for cluster_energy in cluster_energies:
				Current_Population_details_TXT.write(str(cluster_energy) + "\t")
			Current_Population_details_TXT.write("\n")

	def get_current_generation_from_state_file(self):
		"""
		This method is used to get the current generation from the Organisms program if it had already run previously.
		This method is used at the beginnning of the Organisms program to get the current generation if the Organisms program has been
		restarted.

		Returns:
			The current generation of the Organisms program if it has already run previously and is being restarted.
			The names of the clusters in the population at that generation.
		"""
		# looking into the state file for information. 
		if os.path.exists(self.current_population_details_path+'.backup') and self.is_data_in_current_state_file(self.current_population_details_path+'.backup'):
			current_generation, cluster_names, cluster_energies = self.get_data_from_current_state_file(self.current_population_details_path+'.backup')
			return current_generation, cluster_names, cluster_energies, True
		if os.path.exists(self.current_population_details_path)           and self.is_data_in_current_state_file(self.current_population_details_path):
			current_generation, cluster_names, cluster_energies = self.get_data_from_current_state_file(self.current_population_details_path)
			return current_generation, cluster_names, cluster_energies, False
		
		# Could not find the state file or the backup state file. Will attempt to get the relavant data from the history file. 
		if self.write_collection_history:
			self.open('r')
			cluster_energies = []; cluster_names = []; current_generation = None
			# get in PoolProfile.txt the last generation that successfully ran
			for line in reversed(self.history_file.readlines()):
				# get the energies of the clusters of last generation.
				if 'Energies of Clusters:' in line:
					try:
						cluster_energies = line.split()[3:]
						cluster_energies = [float(x) for x in cluster_energies]
					except:
						cluster_energies = None
				# get the dirs of the clusters of last generation.
				elif 'Clusters in Pool:' in line:
					try:
						cluster_names = line.split()[3:]
						cluster_names = [int(x.split('(')[0]) for x in cluster_names]
					except:
						cluster_names = None
				# get the last generation that was successfully completed before the genetic algorithm fail.
				elif 'GA Iteration:' in line:
					try:
						current_generation = line.replace('GA Iteration: ','')
						current_generation = int(current_generation.replace('\n',''))
					except:
						current_generation = None
					if (not cluster_energies == None and not cluster_names == None and not current_generation == None) and (len(cluster_names) == self.size and len(cluster_energies) == self.size):
						break
			self.close()
			#clusters = [(cluster_name, cluster_energy) for cluster_name, cluster_energy in zip(cluster_names,cluster_energies)]
			if current_generation == None:
				self.delete_collection_database()
			self.repair_current_state_file(current_generation, cluster_names, cluster_energies)
			return current_generation, cluster_names, cluster_energies, None
		print('------------------------------------')
		print('ERROR: You do not have any of the following files, or some of the following files are not complete: ')
		print('--> '+self.current_population_details_path+'.backup')
		print('--> '+self.current_population_details_path)
		print('--> '+self.history_path)
		print('Without any of these files, the genetic algorithm can not verify the current generation')
		print()
		print('--> If you have the information, you could consider making a '+str(self.current_population_details_path)+' file yourself in the format')
		print()
		print('GA Iteration: XXX')
		print('Clusters in Pool: XXX XXX XXX ...')
		print('Energies of Clusters: XXX.XX XXX.XX XXX.XX ...')
		print()
		print('Where the "GA Iteration" is the generation that the GA is up to, "Clusters in Pool" are the names of the clusters in the population, and "Energies of Clusters" are the energies of those cluster in the correct units.')
		print()
		print('Check this out')
		exit('This program will exit without completing')

	# ----------------------------------------------------------------------------------------------------------------------------------------------------------
	# Debugging printing information

	def get_details(self):
		"""
		Debugging tool for the Population Class.

		This definition is designed to print all the details about the population.

		"""
		details = [['population_name',self.name],['size',self.size],['Size of Clusters_In_Pop',self.size]]
		print('-----------------------------------------')
		print('Debugging details of the population: '+str(self.name))
		for dd in details:
			print(dd[0]+' = '+str(dd[1]))
		print('-----------------------------------------')

	def print_clusters(self):
		"""
		Debugging tool for the Population Class.

		This definition is designed to print all the details about the clusters in the population.
		
		"""
		print('-----------------------------------------')
		print('Clusters in this population: '+str(self.name))
		highest_energy = -float('inf')
		lowest_energy = float('inf')
		index_highest_energy = -1.1
		index_lowest_energy = -1.1
		print('Sorted by position in the population')
		for index in range(len(self)):
			if 'fitness' in self[index].__dict__.keys():
				print('pop index: ' + str(index) + '; cluster name: ' + str(self[index].name) + '; cluster energy: ' + str(self[index].energy) + ' eV; cluster fitness: ' + str(self[index].fitness) + '')
			else:
				print('pop index: ' + str(index) + '; cluster name: ' + str(self[index].name) + '; cluster energy: ' + str(self[index].energy) + ' eV')
			if self[index].energy > highest_energy:
				highest_energy = self[index].energy
				index_highest_energy = index
			if self[index].energy < lowest_energy:
				lowest_energy = self[index].energy
				index_lowest_energy = index
		print('Sorted by Energy (from lowest to highest energy).')
		indices_of_pop = range(len(self))
		indices_of_pop_sorted = []; sorted_pop_by_energy = [];
		for x,y in sorted(zip(self,indices_of_pop), key = lambda x: x[0].energy, reverse=False):
			sorted_pop_by_energy .append(x)
			indices_of_pop_sorted.append(y)
		for index in range(len(sorted_pop_by_energy)):
			if 'fitness' in sorted_pop_by_energy[index].__dict__.keys():
				print('pop index: ' + str(indices_of_pop_sorted[index]) + '; cluster name: ' + str(sorted_pop_by_energy[index].name) + '; cluster energy: ' + str(sorted_pop_by_energy[index].energy) + ' eV; cluster fitness: ' + str(sorted_pop_by_energy[index].fitness) + '')
			else:
				print('pop index: ' + str(indices_of_pop_sorted[index]) + '; cluster name: ' + str(sorted_pop_by_energy[index].name) + '; cluster energy: ' + str(sorted_pop_by_energy[index].energy) + ' eV')
		print('-----------------------------------------')





class Tree(dict): #WeakValueDictionary()
	"""
	This is a Tree designed for the CNA_Database to hold references of CNA_Entry.
	"""
	def __missing__(self, key):
		"""
		This rewrites the __missing__ component of a dict to work as a tree.
		"""
		value = self[key] = type(self)()
		return value

class LoD_Comparison_Database:
	"""
	This is a database that holds all the entries of CNA_Entry objects.
	"""
	def __init__(self):
		self.database = Tree()

	def __len__(self):
		"""
		Get the number of clusters in the database.
		"""
		return len(self.database)

	def __getitem__(self,cluster_dir):
		"""
		This def will return the branch of the Tree database you want to access.

		:param cluster_name: The is the name of the cluster you would like to obtain CNA Diversity information about.
		:type  cluster_name: int

		:returns: A branch of a tree. 
		:rtypes: Tree branch (dict) 

		"""
		tree_branch = self.database[cluster_dir] # This is new, check out if needed or if it breaks the algorithm from working.
		if tree_branch == {}:
			del self.database[cluster_dir]
			print('Error in def __getitem__ of Class CNA_Database of Diversity_Scheme.py')
			print('Something is not correct with inputs cluster_dir')
			print('This entry does not yet exist in self.CNA_database')
			print('cluster_dir = '+str(cluster_dir))
			print('Check this out')
			import pdb; pdb.set_trace()
			exit()
		return tree_branch

	def reset(self):
		"""
		Reset the LoD database with a new database
		"""
		del self.database
		self.database = Tree()

	def keys(self):
		"""
		give the names of the clusters in the database

		:returns: The list of the names of the clusters in the database. 
		:rtype:   list of ints

		"""
		return self.database.keys()

	def get_cluster_names(self,order=False):
		"""
		Will provide a list of all the names of all the clusters in the Collection

		:param order: This tag will tell this method whether the user would like the list of names given in order. 
		:type  order: bool

		Returns:
			List of the names of all the clusters in the Population
		"""
		cluster_names = list(self.keys())
		if order == True:
			cluster_names.sort()
		return cluster_names

	def get_entry(self,dir_1,dir_2):
		"""
		This def will return the CNA results from the comparison of two clusters.

		:param name_1: the name of the first cluster you want to compare with.
		:type  name_1: int
		:param name_2: the name of the second cluster you want to compare with.
		:type  name_2: int

		:returns: The CNA_Entry that contains all the CNA information about the comparison of these two clusters.
		:rtypes: CNA_Entry

		"""
		CNA_entry_1 = self.database[dir_1][dir_2]; CNA_entry_2 = self.database[dir_1][dir_2]
		if CNA_entry_1 == {} or CNA_entry_2 == {}:
			print('Error in class CNA_Database, in CNA_Database.py')
			print('There is no entry in the CNA_Database between '+str(dir_1)+' and '+str(dir_2))
			print('dir_1 = '+str(dir_1))
			print('dir_2 = '+str(dir_2))
			print('self.database['+str(dir_1)+']['+str(dir_2)+'] = '+str(self.database[dir_1][dir_2]))
			print('self.database['+str(dir_2)+']['+str(dir_1)+'] = '+str(self.database[dir_2][dir_1]))
			print('Check this.')
			import pdb; pdb.set_trace()
			exit()
		elif not CNA_entry_1 is CNA_entry_2:
			print('Error in class CNA_Database, in CNA_Database.py')
			print('self.database[dir_1][dir_2] is not the same as self.database[dir_2][dir_1]')
			print('dir_1 = '+str(dir_1))
			print('dir_2 = '+str(dir_2))
			print('self.database['+str(dir_1)+']['+str(dir_2)+'] = '+str(self.database[dir_1][dir_2]))
			print('self.database['+str(dir_2)+']['+str(dir_1)+'] = '+str(self.database[dir_2][dir_1]))
			print('This should not be the case. What should be true is self.database[dir_1][dir_2] == self.database[dir_2][dir_1] == CNA_Entry(cluster1,cluster2).')
			print('Check this.')
			import pdb; pdb.set_trace()
			exit()
		return CNA_entry_1

	def add(self, dir_1, dir_2, result):
		"""
		This def will obtain the CNA Information for the comparison of name_1 and name_2, and add it to the CNA_Database

		:param name_1: The name of the first cluster to compare against.
		:type  name_1: int
		:param name_2: The name of the second cluster to compare against.
		:type  name_2: int
		:param result: The result of the comparison of the IDCM. True if they are structurally similar by the IDCM, False if not.
		:type  result: bool

		"""
		# Get the CNA comparisons of the two cluters
		if dir_1 == dir_2:
			print('Error in def add of Class CNA_Database of Diversity_Scheme.py')
			print('cluster_1 and cluster_2 have the same dir')
			print('cluster_1.dir = '+str(dir_1))
			print('cluster_2.dir = '+str(dir_2))
			print('Check this out')
			import pdb; pdb.set_trace()
			exit()
		elif (not self.database[dir_1][dir_2] == {}) and (not self.database[dir_2][dir_1] == {}) and (not self.database[dir_1][dir_2] is self.database[dir_2][dir_1]):
			print('Error in class CNA_Database, in CNA_Database.py')
			print('self.database[dir_1][dir_2] is not the same as self.database[dir_2][dir_1]')
			print('dir_1 = '+str(dir_1))
			print('dir_2 = '+str(dir_2))
			print('self.database['+str(dir_1)+']['+str(dir_2)+'] = '+str(self.database[dir_1][dir_2]))
			print('self.database['+str(dir_2)+']['+str(dir_1)+'] = '+str(self.database[dir_2][dir_1]))
			print('This should not be the case. What should be true is self.database[dir_1][dir_2] == self.database[dir_2][dir_1] == CNA_Entry(cluster1,cluster2).')
			print('Check this.')
			import pdb; pdb.set_trace()
			exit()
		#new_entry = CNA_Entry(dir_1,CNA_profile_dir_1,dir_2,CNA_profile_dir_2,self.rCut_low,self.rCut_high,self.rCut_resolution,self.get_CNA_similarities_method,self.get_comparison_method)
		self.database[dir_1][dir_2] = result
		self.database[dir_2][dir_1] = result

	def remove(self,dir_to_remove):
		"""
		Remove all entries that exists in the database that are associated with a particular cluster.

		:param name_to_remove: The name of the cluster you would like to remove from the CNA_Database.
		:type  name_to_remove: int

		"""
		LoD_entries_for_other_dirs = self.database[dir_to_remove].keys()
		# delete entries that include dir_to_remove in all other branches in the CNA_Database
		for LoD_entries_for_other_dir in LoD_entries_for_other_dirs:
			del self.database[LoD_entries_for_other_dir][dir_to_remove]
		# remove the complete branch associated with dir_to_remove
		del self.database[dir_to_remove]

	def which_clusters_in_LoD_comparison_database_are_similar(self):
		"""
		This method will return a dict of all the similar clusters in the LoD database, where pairs of clusters have been deemed structurally similar by the IDCM. 

		:returns: A dictionary of all the clusters that have been deemed structurally similar to each other. The format of this dictionary is {name_1: [name_2, name_3, ..., names of all the clusters that name_1 is structurally similar to by the IDCM], ...}. 
		:rtypes: dict. 

		"""
		identical_pairs = {}
		clusters_in_database = list(self.keys())
		# This will make a 2D matrix (as a dict) of similarity data.
		for index1 in range(len(clusters_in_database)):
			name1 = clusters_in_database[index1]
			for index2 in range(index1+1,len(clusters_in_database)):
				name2 = clusters_in_database[index2]
				if self[name1][name2]:
					identical_pairs.setdefault(name1,[]).append(name2)
					identical_pairs.setdefault(name2,[]).append(name1)
		return identical_pairs

	def is_cluster_pair_in_the_database(self,dir_1,dir_2):
		"""
		Determine if a CNA entry exists for two clusters in the database

		:param name_1: The name of the first cluster you would like to look for an entry in the CNA_Database.
		:type  name_1: int
		:param name_2: The name of the second cluster you would like to look for an entry in the CNA_Database.
		:type  name_2: int

		:returns: True if this cluster pair is in the database, False if not. 
		:rtypes: bool

		"""
		if self.database[dir_1][dir_2] == {}:
			if not self.database[dir_2][dir_1] == {}:
				print('Error in class CNA_Database, in CNA_Database.py')
				print('The Tree is not full')
				print('There is an entry for self.database[dir_1][dir_2], but not for self.database[dir_2][dir_1].')
				print('dir_1 = '+str(dir_1))
				print('dir_2 = '+str(dir_2))
				print('self.database['+str(dir_1)+']['+str(dir_2)+'] = '+str(self.database[dir_1][dir_2]))
				print('self.database['+str(dir_2)+']['+str(dir_1)+'] = '+str(self.database[dir_2][dir_1]))
				print('This should not be the case. What should be true is self.database[dir_1][dir_2] == self.database[dir_2][dir_1] == CNA_Entry(cluster1,cluster2).')
				print('Check this.')
				import pdb; pdb.set_trace()
				exit()
			del self.database[dir_2][dir_1]
			del self.database[dir_1][dir_2]
			return False
		else:
			if self.database[dir_2][dir_1] == {}:
				print('Error in class CNA_Database, in CNA_Database.py')
				print('The Tree is not full')
				print('There is an entry for self.database[dir_1][dir_2], but not for self.database[dir_2][dir_1].')
				print('dir_1 = '+str(dir_1))
				print('dir_2 = '+str(dir_2))
				print('self.database['+str(dir_1)+']['+str(dir_2)+'] = '+str(self.database[dir_1][dir_2]))
				print('self.database['+str(dir_2)+']['+str(dir_1)+'] = '+str(self.database[dir_2][dir_1]))
				print('This should not be the case. What should be true is self.database[dir_1][dir_2] == self.database[dir_2][dir_1] == CNA_Entry(cluster1,cluster2).')
				print('Check this.')
				import pdb; pdb.set_trace()
				exit()
			elif not self.database[dir_1][dir_2] is self.database[dir_2][dir_1]:
				print('Error in class CNA_Database, in CNA_Database.py')
				print('self.database[dir_1][dir_2] is not the same as self.database[dir_2][dir_1]')
				print('dir_1 = '+str(dir_1))
				print('dir_2 = '+str(dir_2))
				print('self.database['+str(dir_1)+']['+str(dir_2)+'] = '+str(self.database[dir_1][dir_2]))
				print('self.database['+str(dir_2)+']['+str(dir_1)+'] = '+str(self.database[dir_2][dir_1]))
				print('This should not be the case. What should be true is self.database[dir_1][dir_2] == self.database[dir_2][dir_1] == CNA_Entry(cluster1,cluster2).')
				print('Check this.')
				import pdb; pdb.set_trace()
				exit()
			return True

	def LoD_Similarity_Analysis(self,dir_1,dir_2):
		"""
		This def will obtain the max similarity percentage from the compararison of two clusters, name_1 and name_2.

		:param name_1: The name of the first cluster you would like to look for an entry in the CNA_Database.
		:type  name_1: int
		:param name_2: The name of the second cluster you would like to look for an entry in the CNA_Database.
		:type  name_2: int

		:returns: The result from the IDCM. This will be True if this cluster pair is in the database, False if not. 
		:rtypes: bool

		"""
		LoD_entry = self.get_entry(dir_1,dir_2)
		return LoD_entry
		
	def are_all_entries_false(self,all_collections):
		"""
		This method checks if all entries in the database are false. This method performs this after clusters have been removed from the population or offspring_pool due to violating the IDCM predatino operator. At this point, all entries should be false. 
		
		:param all_collections: This is a list of the population and the offspring pool.
		:type  all_collections: list of Organisms.GA.Collections.Collections

		"""
		for dir1, tree_branch in self.database.items():
			for dir2, similarity_result in tree_branch.items(): 
				if similarity_result == True:
					print('Error in LoD_Comparison_Database')
					print('Found two identical clusters in database.')
					print('This should not be the case at this point in the GA.')
					print('cluster1: '+str(dir1)+', cluster2: '+str(dir2))
					print('------------------------------------------------------------')
					self.make_database_table()
					print('------------------------------------------------------------')
					print('Check this')
					import pdb; pdb.set_trace()
					exit('This program will exit without completing')
		size_of_all_collections = sum([len(a) for a in all_collections])
		if not len(self.database.keys()) == size_of_all_collections:
			print('Error1')
			import pdb; pdb.set_trace()
			exit()

		for dir1, tree_branch in self.database.items():
			if not len(tree_branch) == size_of_all_collections - 1:
				print('Error2')
				import pdb; pdb.set_trace()
				exit()

	def check_for_issues(self, population):
		"""
		This method check to make sure there are no issue with the LoD Database after the natural selection process

		:param population: This is the population
		:type  population: Organisms.GA.Population.Population
		"""
		all_cluster_names = population.get_cluster_names(order=True)
		all_cluster_names_in_database = self.get_cluster_names(order=True)
		if not all_cluster_names == all_cluster_names_in_database: 
			print("The lists are not identical, check_for_issues, LoD Database") 
			import pdb; pdb.set_trace()
			exit()

		for dir1, tree_branch in self.database.items():
			for dir2, similarity_result in tree_branch.items(): 
				if similarity_result == True:
					print('Error in LoD_Comparison_Database')
					print('During Generation, No Epoch')
					print('Found two identical clusters in database.')
					print('This should not be the case at this point in the GA.')
					print('cluster1: '+str(dir1)+', cluster2: '+str(dir2))
					print('------------------------------------------------------------')
					#self.make_database_table()
					print('------------------------------------------------------------')
					print('Check this')
					import pdb; pdb.set_trace()
					exit('This program will exit without completing')

	def make_database_table(self):
		"""
		This method prints a table of the database. 
		"""
		all_cluster_names_in_database = self.get_cluster_names(order=True)
		row_format ="{:>7}" * (len(all_cluster_names_in_database) + 1)
		print(row_format.format("", *all_cluster_names_in_database))
		for cluster_name1 in all_cluster_names_in_database:
			similarities = []
			for cluster_name2 in all_cluster_names_in_database: 
				if cluster_name1 == cluster_name2:
					similarity = '--'
				else:
					similarity = str(self.database[cluster_name1][cluster_name2])
				similarities.append(similarity)
			#import pdb; pdb.set_trace()
			print(row_format.format(cluster_name1, *similarities))













'''
Did_Find_LES.py, Geoffrey Weal, 08/03/2019

This program will determine which of your genetic algorithm trials have completed up to a certain generation. 
'''
import os, sys
import subprocess

def tail(f, n, offset=0):
    proc = subprocess.Popen(['tail', '-n', str(n + offset), f], stdout=subprocess.PIPE)
    lines = proc.stdout.readlines()
    return lines

def get_variables_from_run(filepath):
    with open(filepath+'/Run.py','r') as RunPY:
        generations = None; no_offspring_per_generation = None;
        for line in RunPY:
            if line.startswith('generations = '):
                generations = int(eval(line.replace('generations = ','')))
                if (generations == None and no_offspring_per_generation == None):
                    break
            if line.startswith('no_offspring_per_generation = '):
                no_offspring_per_generation = int(eval(line.replace('no_offspring_per_generation = ','')))
                if (generations == None and no_offspring_per_generation == None):
                    break
    if generations == None or no_offspring_per_generation == None:
        print('Error')
        print(generations)
        print(no_offspring_per_generation)
        import pdb; pdb.set_trace()
        exit('Error')
    return generations, no_offspring_per_generation

def Did_Trial_finish_successfully(filepath):
    # get decired generations and number of offspring per gerenation
    total_no_of_generation, no_offspring_per_generation = get_variables_from_run(filepath)
    # Read EnergyProfile.txt
    #with open(filepath+'/Population/EnergyProfile.txt','r') as EnergyProfileTXT:
    last_lines_in_EnergyProfile = tail(filepath+'/Population/EnergyProfile.txt',no_offspring_per_generation) 
    all_cluster_gen_made = []
    finished_found_LES = False
    Restart_due_to_epoch = False
    for line in last_lines_in_EnergyProfile:
        if line.startswith('Finished prematurely as LES energy found.'):
            finished_found_LES = True
        elif line.startswith('Restarting due to epoch.'):
            Restart_due_to_epoch = True
        else:
            line = line.split()
            generation = int(line[1])
            all_cluster_gen_made.append(generation)
    if finished_found_LES:
        return True, all_cluster_gen_made[0]
    if Restart_due_to_epoch:
        return False, all_cluster_gen_made[0]
    if not (len(set(all_cluster_gen_made)) == 1):
        print('1')
        import pdb; pdb.set_trace()
        exit('Error')
    all_cluster_gen_made = all_cluster_gen_made[0]
    if all_cluster_gen_made < total_no_of_generation:
        return False, all_cluster_gen_made
    elif all_cluster_gen_made == total_no_of_generation:
        return True, all_cluster_gen_made
    else:
        #print('2')
        #import pdb; pdb.set_trace()
        #exit('Error')
        print('Note, you have performed more than enough generations for: '+str(filepath))
        return True, all_cluster_gen_made

def has_all_trials_finished(dirpath, dirnames):

    completed_Trials = []
    incomplete_Trials = []
    at_generation = []
    dirnames = [dirname for dirname in dirnames if dirname.startswith('Trial')]
    dirnames.sort(key=lambda dirname: int(dirname.replace('Trial','')))
    number_of_dirnames = len(dirnames)
    for index in range(number_of_dirnames):
        dirname = dirnames[index]
        trial_no = int(dirname.replace('Trial',''))
        try:
            did_trial_finish_successfully, current_gen = Did_Trial_finish_successfully(dirpath+'/'+dirname)
        except Exception as expection:
            print('===================================================================')
            print('Error in def has_all_trials_finished in Did_Complete_Main.py')
            print('Something weird is happening with Trial '+str(trial_no))
            print('Getting the following expection:')
            print()
            print('------------->')
            print(expection)
            print('------------->')
            print()
            print('Check this out and then repeat your program again')
            print('This program will finish here.')
            exit ('===================================================================')
        if did_trial_finish_successfully:
            completed_Trials.append(trial_no)
        else:
            incomplete_Trials.append(trial_no)
            at_generation.append((dirname,current_gen))
        ###########
        sys.stdout.write("\r                                                                  ")
        sys.stdout.flush()
        sys.stdout.write("\rScanning Completion: "+str(float(index+1)/float(number_of_dirnames)*100.0)+" % (Checked "+str(dirname)+").")
        sys.stdout.flush()
        ###########
    completed_Trials.sort()
    incomplete_Trials.sort()
    at_generation.sort(key=lambda x:x[1])
    completed_successfully = True if len(incomplete_Trials) == 0 else False
    return completed_successfully, completed_Trials, incomplete_Trials, at_generation



'''
Cluster.py, 12/04/2017, Geoffrey R Weal

This class is specifically designed to hold cluster objects for the 
Genetic Algorithm program. This object will hold a reference to the 
Atoms class from ASE for the cluster, the energy of the cluster and 
the dir tag for the cluster

'''

from ase import Atoms
from copy import deepcopy
from ase.visualize import view
from collections import Counter
from numpy import array, zeros, sort
from time import time

from Organisms.GA.ExternalDefinitions import InclusionRadiusOfCluster

def import_surface(surface):
	"""
	This method is designed to import the surface into the genetic algorithm

	:param surface: This is the surface that the cluster is optimised upon. 
	:type  surface: str/ASE.Atoms.
	"""
	if isinstance(surface, test_Atoms):
		surface = surface
	elif isinstance(surface, str):
		surface = ase_read(surface)
	elif surface == None:
		surface = None
	else:
		exit('Error')
	for index in range(len(surface)):
		surface[index].type = 'surface'

	surface.centre()
	return surface

class Cluster(Atoms):
	'''
	This class is designed to clude all the information that the genetic algorithm needs to know from the chemical system.

	:param symbols: This can be a list of all the chemical symbols of atoms in the clusters, OR, the ASE.Atoms object that this object will be base on. For writing this as a list of element symbols, see See https://wiki.fysik.dtu.dk/ase/ase/atoms.html?highlight=atoms#ase.Atoms for more information. Default: None
	:type  symbols: str (formula) or list of str., or ASE.Atoms
	:param positions: A list of all the (x,y,z) values for the cluster. See https://wiki.fysik.dtu.dk/ase/ase/atoms.html?highlight=atoms#ase.Atoms for more information.  Default: None
	:type  positions: list of xyz-positions
	:param numbers: See https://wiki.fysik.dtu.dk/ase/ase/atoms.html?highlight=atoms#ase.Atoms for more information. Default: None
	:type  numbers: list of int
	:param tags: See https://wiki.fysik.dtu.dk/ase/ase/atoms.html?highlight=atoms#ase.Atoms for more information. Default: None
	:type  tags: list of int
	:param momenta: See https://wiki.fysik.dtu.dk/ase/ase/atoms.html?highlight=atoms#ase.Atoms for more information. Default: None
	:type  momenta: list of xyz-momenta
	:param masses: See https://wiki.fysik.dtu.dk/ase/ase/atoms.html?highlight=atoms#ase.Atoms for more information. Default: None
	:type  masses: list of float
	:param magmoms: See https://wiki.fysik.dtu.dk/ase/ase/atoms.html?highlight=atoms#ase.Atoms for more information. Default: None
	:type  magmoms: list of float or list of xyz-values
	:param charges: See https://wiki.fysik.dtu.dk/ase/ase/atoms.html?highlight=atoms#ase.Atoms for more information. Default: None
	:type  charges: list of float
	:param scaled_positions: See https://wiki.fysik.dtu.dk/ase/ase/atoms.html?highlight=atoms#ase.Atoms for more information. Default: None
	:type  scaled_positions: list of scaled-positions
	:param cell: See https://wiki.fysik.dtu.dk/ase/ase/atoms.html?highlight=atoms#ase.Atoms for more information. Default: None
	:type  cell: 3x3 matrix or length 3 or 6 vector
	:param pbc: See https://wiki.fysik.dtu.dk/ase/ase/atoms.html?highlight=atoms#ase.Atoms for more information. Default: None
	:type  pbc: one or three bool
	:param celldisp: See https://wiki.fysik.dtu.dk/ase/ase/atoms.html?highlight=atoms#ase.Atoms for more information. Default: None
	:type  celldisp: Vector
	:param constraint: See https://wiki.fysik.dtu.dk/ase/ase/atoms.html?highlight=atoms#ase.Atoms for more information. Default: None
	:type  constraint: constraint object(s)
	:param calculator: See https://wiki.fysik.dtu.dk/ase/ase/atoms.html?highlight=atoms#ase.Atoms for more information. Default: None
	:type  calculator: calculator object
	:param info: See https://wiki.fysik.dtu.dk/ase/ase/atoms.html?highlight=atoms#ase.Atoms for more information. Default: None
	:type  info: dict of key-value pairs
	:param surface: This is the surface that the cluster is modelled on. This is given either as a string or the ASE.Atoms object. Default: None
	:type  surface: ASE.Atoms/str.

	'''
	def __init__(self,symbols=None,positions=None,numbers=None,tags=None,momenta=None,masses=None,magmoms=None,charges=None,scaled_positions=None,cell=None,pbc=None,celldisp=None,constraint=None,calculator=None,info=None,surface=None):
		if isinstance(symbols,Atoms):
			positions = symbols.get_positions()
			numbers = symbols.get_atomic_numbers()
			tags = symbols.get_tags()
			momenta = symbols.get_momenta()
			masses = symbols.get_masses()
			magmoms = symbols.get_initial_magnetic_moments()
			charges = symbols.get_initial_charges()
			scaled_positions = None #symbols.get_scaled_positions()
			cell = symbols.get_cell()
			pbc = symbols.get_pbc()
			celldisp = symbols.get_celldisp()
			constraint = [c.copy() for c in symbols.constraints]
			calculator = symbols.get_calculator()
			info = deepcopy(symbols.info)
			symbols = None
		super().__init__(symbols,positions,numbers,tags,momenta,masses,magmoms,charges,scaled_positions,cell,pbc,celldisp,constraint,calculator,info)
		self.verified = False
		self.name = None

	def __str__(self):
		"""
		:returns: a label for this cluster, which is the dir of the cluster
		:rtype: str
		"""
		return 'Cluster_'+str(self.name)

	def __repr__(self):
		"""
		:returns: a label for this cluster, which is the dir of the cluster
		:rtype: str
		"""
		return str(self)

	def view(self):
		"""
		Allow the user to visually look at the cluster using the ASE gui. This is a debugging method.
		"""
		view(self)

	def verify_cluster(self,name,gen_made,vacuum_length,rounding_criteria):
		"""
		This method will verify that the cluster contains all the other information that it needs to run during the genetic algorithm, providing the name, generation made, and the energy of the cluster, as well as centering in a unit cell. 
		This method is performed after the cluster has been locally optimised. 

		:param name: The name that the cluster is referenced in this genetic algorithm program. This should be a integer that is based on when the cluster was made. 
		:type  name: int/str.
		:param gen_made: This is the generation when the cluster was created
		:type  gen_made: int
		:param vacuum_length: This is the amount of vacuum to give the cluster.
		:type  vacuum_length: float
		:param rounding_criteria: The number of decimal places that the cluster's energy is rounded to.
		:type  rounding_criteria: int

		"""
		self.verified = True
		self.name = name
		self.gen_made = gen_made
		self.sorted_by_Z = False
		if not len(self) == 0:
			#energy_start = time()
			self.energy = self.get_total_cluster_energy(rounding_criteria)
			#energy_end = time()
			#print(' '.join(['energy time = ',str(energy_end-energy_start)]))
			self.centre_cluster_at_centre_of_cell(vacuum_length)
		elif self.dir == -1:
			pass
		else:
			print('Error in Cluster, in Cluster.py')
			print('This cluster contains no atoms.')
			print("This probably shouldn't happen")
			print('Check this.')
			import pdb; pdb.set_trace()
			exit('This program will finish without completing.')

	def custom_verify_cluster(self,name,gen_made,cluster_energy,ever_in_population,excluded_because_violates_predation_operator,initial_population):
		"""
		This method allows you to custom verify the cluster. To be used when resuming the genetic algorithm. 

		:param name: The name that the cluster is referenced in this genetic algorithm program. This should be a integer that is based on when the cluster was made. 
		:type  name: int/str.
		:param gen_made: This is the generation when the cluster was created
		:type  gen_made: int
		:param ever_in_population: Was this cluster ever in the population. 
		:type  ever_in_population: bool.
		:param excluded_because_violates_predation_operator: Was the cluster removed from the offspring pool because it violated the predation operator
		:type  excluded_because_violates_predation_operator: bool.
		:param initial_population: Was the cluster apart of the initial population, either at rhe beginning of the population or after an epoch.
		:type  initial_population: bool.

		"""
		self.verified = True
		self.name = name
		self.gen_made = gen_made
		self.sorted_by_Z = False
		self.energy = cluster_energy
		self.ever_in_population = ever_in_population
		self.excluded_because_violates_predation_operator = excluded_because_violates_predation_operator
		self.initial_population = initial_population

	def sortZ(self):
		"""
		Sort the cluster from most positive to most negative value of z. This sorting by z axis is required to allow the algorithm to easily split itself into two sides of a dividing plane.

		"""
		if not self.sorted_by_Z:
			new_cluster = zeros((len(self),4))
			new_cluster.T[0] = self.get_atomic_numbers()
			new_cluster.T[1:] = self.get_positions().T #arrays['positions'].T
			new_cluster = array(sorted(new_cluster,key=lambda atom: atom[3],reverse=True))
			self.set_positions(new_cluster.T[1:].T)
			self.set_atomic_numbers(new_cluster.T[0].T)
			self.sorted_by_Z = True

	def get_total_cluster_energy(self,rounding_criteria=None):	
		'''
		Obtain the energy of the cluster to a predefined decimal place.
		
		:param rounding_criteria: The number of decimal places that the cluster's energy is rounded to.
		:type  rounding_criteria: int

		returns energy: This is the energy of the cluster
		rtype   energy: float
		'''
		energy = self.get_potential_energy()
		if not rounding_criteria == None:
			energy = round(energy,rounding_criteria)
		return energy

	def centre_cluster_at_centre_of_cell(self,vacuumAdd):
		"""
		This method will center the cluster in the middle of the cell. This method has been employed to make it easier to look at clusters that are made during the Organisms program. 

		Inputs:
			vacuumAdd (float): The amount of value to add around the cluster.
		"""
		lengthOfCell = 2.0*InclusionRadiusOfCluster(self) + vacuumAdd
		cell = [lengthOfCell,lengthOfCell,lengthOfCell]
		self.set_cell(cell)
		#start_time = time()
		self.center(vacuum=None, axis=(0, 1, 2), about=0.) # make a better center function?
		#end_time = time()
		#print('center time = '+str(end_time-start_time))

	def get_elemental_makeup(self):
		"""
		This gives a list which indicates the types of elements, and the number of those elements, in the cluster

		:returns: A list which indicates the types of elements, and the number of those elements, in the cluster
		:rtype: {str: int, ...} (old output was [[str.,int],...])
		"""
		return Counter(self.get_chemical_symbols())
		#self.chemical_makeup = Counter(self.get_chemical_symbols())
		#return self.chemical_makeup #Counter(self.get_chemical_symbols())

	def remove_calculator(self):
		"""
		This method will remove the calculator for this Cluster
		"""
		self.set_calculator(calc=None)

	def deepcopy(self):
		"""
		Will create a copy of the Cluster. This copy does not include a copy of the calculator as this has the potential to cause issues.
		"""
		#return self.copy()
		original_calculator = self.get_calculator()
		self.remove_calculator()
		new_Cluster = deepcopy(self)
		self.set_calculator(calc=original_calculator)
		return new_Cluster

	def deepcopy_skeleton(self):
		"""
		This will create a copy of the cluster that does not contain any atoms, but will contain all the other settings for this cluster.
		"""
		skeleton = self.deepcopy()
		for NOTUSED in range(len(skeleton)):
			del skeleton[-1]
		return skeleton


from Organisms.GA.Lock import Lock_Remove

def Check_Population_against_predation_operator(population,predation_operator):
	"""
	Checks that all the clusters in this newly initialised population obey the predation operator

	:param population: The population
	:type  population: Organisms.GA.Population
	:param predation_operator: This is the predation operator 
	:type  predation_operator: Organisms.GA.Predation_Operator

	:returns a list of all the clusters that do not obey the predation operator, as well as a report about the issues.
	:rtype   A list of int, and a string
	"""
	# The population will be checked by the Fitness_Factor_Scheme for duplicates. Check_Initial_Population() will delete any clusters that are repeats and present them as cluster_numbers_to_create
	clusters_to_create, removed_clusters_report = predation_operator.check_initial_population(return_report=True)
	# Check to see the number of entries in clusters_to_create is as expected.
	if not population.size-len(population) == len(clusters_to_create):
		print('Error def Initate_New_GAProgram of GAProgram.py: Something weird is going on. The following command: ')
		print('population.size-len(population) == len(clusters_to_create) = False')
		print('It is expected that the difference in the population compared to the maximum number that we want should equal the number of entried in cluster_numbers_to_create')
		print('Not too sure what is going on. Check this.')
		print('population.size = '+str(population.size))
		print('len(population) = '+str(len(population)))
		print('population.size-len(population) = '+str(population.size-len(population)))
		print('len(clusters_to_create) = '+str(len(clusters_to_create)))
		import pdb; pdb.set_trace()
		exit('Program will finish with out completing')

	if len(population) > population.size: 
		print('Error def Initate_New_GAProgram of GAProgram.py: Something weird is going on.')
		print('The population is bigger than the set size by pop_size.')
		print('len(self) (The size of this population) = '+str(len(population)))
		print('self.size  (for class Population) = '+str(population.size))
		print('Check this')
		import pdb; pdb.set_trace()
		exit('Program will finish with out completing')
	elif len(population) < population.size: # If any clusters need to be created to fit inline with the Fitness_Factor_Scheme, repeat the process to get new randomised clusters (repeat while loop).
		if not len(clusters_to_create) > 0:
			print('Error def Initate_New_GAProgram of GAProgram.py: Something weird is going on.')
			print('Shoud not have been caught up in this error. Should have been caught up in the previous error.')
			print('Number of clusters to make = '+str(len(clusters_to_create)))
			print('Check this')
			import pdb; pdb.set_trace()
			exit('Program will finish with out completing')
		print('*****************************')
		print('*****************************')
		print('Diversity Scheme located "duplicates" in creating population.')
		print('Recreating clusters: ' + str([x[1] for x in clusters_to_create]))
		return clusters_to_create, removed_clusters_report
	elif len(population) == population.size:
		return clusters_to_create, removed_clusters_report # Finished initialisation process. The pool is complete and filled with self.size random structures.
	else:
		print('Error in GAProgram, in def Initate_New_GAProgram.')
		print('The population is bigger than its maximum size.')
		print('Check this.')
		print('self.size  (for class Population) = '+str(population.size))
		print('len(self) = ' + str(len(population)))
		import pdb; pdb.set_trace()
		exit('Program will finish with out completing')

# ----------------------------------------------------------------------------------------------------------------------------- #
# ----------------------------------------------------------------------------------------------------------------------------- #
# ----------------------------------------------------------------------------------------------------------------------------- #

import os
from ase.io import read as ase_read
from Organisms.GA.Cluster import Cluster
from Organisms.GA.ExternalDefinitions import Exploded
def Place_Already_Created_Clusters_In_Population(population,cluster_makeup,Minimisation_Function,vacuum_to_add_length,r_ij,rounding_criteria,surface,memory_operator,predation_operator,fitness_operator,previous_cluster_name):
	"""
	This method will place any user created clusters into the population.

	:param population: The population
	:type  population: Organisms.GA.Population
	:param cluster_makeup: The makeup of the cluster
	:type  cluster_makeup: dict.
	:param Minimisation_Function: The minimisation function
	:type  Minimisation_Function: __func__
	:param vacuum_to_add_length: The amount of vacuum to place around the cluster
	:type  vacuum_to_add_length: float
	:param r_ij: The maximum distance that should be between atoms to be considered bonded. This value should be as large a possible, to reflected the longest bond possible between atoms in the cluster.
	:type  r_ij: float
	:param rounding_criteria: The number of decimal places to round the energy of clusters to.
	:type  rounding_criteria: int
	:param surface: This is the surface that the cluster is placed on. None means there is no surface.
	:type  surface: Organisms.GA.Surface
	:param memory_operator: The memory operator
	:type  memory_operator: Organisms.GA.Memory_Operator
	:param predation_operator: This is the predation operator 
	:type  predation_operator: Organisms.GA.Predation_Operator
	:param fitness_operator: This is the fitness operator
	:type  fitness_operator: Organisms.GA.Fitness_Operator
	:param previous_cluster_name: This is the name of the last cluster created in the genetic algorithm.
	:type  previous_cluster_name: int

	:returns The name of the most recently created cluster by this method in the population.
	:rtype   int

	"""
	print('Checking to see if the user has any clusters in the folder '+str(population.user_initialised_population_folder)+' that the user would like to have inputted into the initalise population.')
	# First, the definition will check the clusters if there is a population folder. If not, the user does not want to add any clusters to the initial population.
	if population.user_initialised_population_folder == None or population.user_initialised_population_folder == '':
		return 0
	if not os.path.exists(population.user_initialised_population_folder):
		print('Error in def Place_Already_Created_Clusters_In_Population of class Population, in Population.py.')
		print('Yo, The path to the Intialised Population folder that you have specified does not exist.')
		print('Check the path to this folder exists: '+str(population.user_initialised_population_folder))
		print('Note: This variable should include the folder itself as well in the path.')
		exit('This program will end without doing anything.')
	##############################################################################################################################################
	print('################################################################################################################')
	print('################################################################################################################')
	print('Will now initalise the population with clusters from:')
	print(str(population.user_initialised_population_folder))
	print('################################################################################################################')
	print('################################################################################################################')
	##############################################################################################################################################
	# The folder nammed by the variable self.user_initialised_population_folder exists. Lets get whatever clusters are inside this folder.
	cluster_names = [] # holds the cluster dir values, which are equivalent to the number of the folder
	for cluster_name in os.listdir(population.user_initialised_population_folder):
		if cluster_name.startswith('.'):
			continue
		# If the cluster folder exists and the folder is numbered properly, then it is a cluster to be added into the GA program
		if os.path.isdir(os.path.join(population.user_initialised_population_folder, cluster_name)) and cluster_name.isdigit():
			cluster_names.append(int(cluster_name))
		else:
			print('Error in def Place_Already_Created_Clusters_In_Population of class Population, in Population.py.')
			print('Check the Initialised Population folder '+str(population.user_initialised_population_folder)+'. There is a folder which is not an integer')
			print('All folders must be called by an integer as this is how the program names each cluster.')
			print('Check the folder in '+str(population.user_initialised_population_folder)+' to make sure all folders are clusters with integer labels.')
			print('The folder that the GA is specifically having problems with is called '+str(cluster_name))
			print('This program will end without doing anything.')
			exit()
	cluster_names.sort()
	##############################################################################################################################################
	# The clusters that the user has put into the population currently should be in sequential order starting from 1.
	def is_list_sequential(mylist):
		for index in range(0,len(mylist)-1):
		    if not mylist[index+1] == mylist[index]+1:
			    return False
		return True
	if cluster_names == []:
		print('Error in def Place_Already_Created_Clusters_In_Population of class Population, in Population.py.')
		print('There are no clusters in the Initialised Population in the path: '+str(population.user_initialised_population_folder))
		print('Check the Initialised Population folder '+str(population.user_initialised_population_folder))
		print('This program will end without doing anything.')
		exit()
	elif not 1 in cluster_names:
		print('Error in def Place_Already_Created_Clusters_In_Population of class Population, in Population.py.')
		print('One of the clusters in your already created population must have a dir/folder name that is the name "1"')
		print('Check the cluster folders that are in the directory '+str(population.user_initialised_population_folder))
		print('Folder names in '+str(population.user_initialised_population_folder)+' folder: '+str(cluster_names))
		print('This program will end without doing anything.')
		exit()
	elif not is_list_sequential(cluster_names):
		print('Error in def Place_Already_Created_Clusters_In_Population of class Population, in Population.py.')
		print('The names of the clusters/folders in the population folder "'+str(population.user_initialised_population_folder)+'" must be named by numbers in sequential order starting from "1"')
		print('Check the cluster folders that are in the directory '+str(population.user_initialised_population_folder))
		print('Folder names in '+str(population.user_initialised_population_folder)+' folder: '+str(cluster_names))
		print('This program will end without doing anything.')
		exit()
	##############################################################################################################################################
	# The clusters will now be locally minimised as described by the def Minimisation_Function
	for cluster_name in cluster_names:
		##########################################
		# Should be ok, but lets just double check at least that the clusters dir is between 1 and self.size
		if not cluster_name > 0:
			print('Error in def Place_Already_Created_Clusters_In_Population of class Population, in Population.py.')
			print('The dir name of this cluster is less than 0.')
			print('Dir of this cluster: '+str(cluster_name))
			print('Cluster is located in '+str(population.user_initialised_population_folder+'/'+str(cluster_name)))
			print('Check this out')
			import pdb; pdb.set_trace()
			exit()
		elif not cluster_name < population.size+1:
			toStringError = '******************************************************************\n'
			toStringError += 'There are more clusters in '+str(population.user_initialised_population_folder)+' than allowed in this population.\n'
			toStringError += 'The population size for this genetic algorithm run is: '+str(population.size)+'\n'
			toStringError += 'What we are going to do is not import any of the other clusters in the population, only keeping the following clusters\n'
			included_clusters_in_pop = population.get_cluster_names()
			toStringError += str(included_clusters_in_pop)+'\n'
			toStringError += 'We will exclude the following clusters from the population.\n'
			excluded_clusters_in_pop = []
			for cluster_name in cluster_names:
				if not cluster_name in included_clusters_in_pop:
					excluded_clusters_in_pop.append(cluster_name)
			toStringError += str(excluded_clusters_in_pop)+'\n'
			toStringError += 'The genetic algorithm will continue. I hope you know what you are doing.\n'
			toStringError += '******************************************************************\n'
			print(toStringError)
			print(toStringError, file=sys.stderr)
			break
		##########################################
		print('*****************************')
		xyz_files = [file for file in os.listdir(os.path.join(population.user_initialised_population_folder, str(cluster_name))) if file.endswith('.xyz')]
		# Check to find the xyz file that will be used. This xyz file will be considered unoptimsed
		if len(xyz_files) == 0:
			print('Check cluster '+str(cluster_name)+'. There is no .xyz file. The prorgram requires a .xyz file of this cluster so it knows the positions of the atoms in the cluster.')
			exit()
		# File the cluster_name_UnOpt.xyz file. If it cant be found. Tell the user that it doesnt exist 
		# and that this needs to be resolved by them. Note: we want to obtained the "_UnOpt.xyz" file,
		# NOT the _Opt.xyz. This is because we will be optimising it later, and for ease of the programming.
		# I know you might have already use the get_newly initialised_population.py program to get both the 
		# unoptimised and optimised clusters, so you have already optimised the cluster, but this is just easier
		# overall. Take the optimised cluster obtained during the initialised_population.py program as a test to mkae sure 
		# that the unoptimised cluster will not explode before it is used in the GA program. If you want to, you can change
		# this to use the "_Opt.xyz" file, and it probably wont be an issue. You may have to do some changing of the program, 
		# but im sure it will be fine to do.
		for xyz_file in xyz_files:
			if str(cluster_name)+'_UnOpt.xyz' == xyz_file:
				break
		else:
			print('Error in def Place_Already_Created_Clusters_In_Population of class Population, in Population.py.')
			print('There is no .xyz file in Cluster (folder) '+str(cluster_name)+' called " '+str(cluster_name)+'_Opt.xyz.')
			print('This is the cluster xyz file that is used by the genetic algorithm to represent this cluster, as it is assume that it is that cluster, and the "_Opt" indicates that it has been locally optimised.')
			print('Check folder '+str(cluster_name)+' in '+str(population.user_initialised_population_folder)+' for the file "'+str(str(cluster_name)+'_Opt.xyz')+'".')
			print('Ending the genetic algorithm without doing anything.')
			exit()
		# Performed all checks, will now add cluster
		print('Adding Cluster '+str(cluster_name)+' to the population.')
		UnOpt_Cluster = ase_read(population.user_initialised_population_folder+'/'+str(cluster_name)+'/'+xyz_file)
		print('Will locally optimise it as described by your def Minimisation_Function.')
		stdout = sys.stdout; output = StringIO(); sys.stdout = output
		Opt_Cluster, converged, opt_information = Minimisation_Function(UnOpt_Cluster.copy(),population.name,cluster_name) # note, self.name is the population name
		sys.stdout = stdout
		opt_information['output.txt'] = output.getvalue()
		if Exploded(Opt_Cluster,max_distance_between_atoms=r_ij): # make sure the randomised cluster has not split up when optimised.
			print('Error in def Place_Already_Created_Clusters_In_Population of class Population, in Population.py.')
			print("The optimised Cluster exploded. Check out why cluster "+str(cluster_name)+' has exploded by looking at cluster '+str(cluster_name)+'s AfterOpt.traj file.')
			print('The program will now show the optimised and the unoptimised versions of cluster '+str(cluster_name))
			from ase.visualize import view; view(UnOpt_Cluster); view(Opt_Cluster)
			print('This program will end without doing anything.')
			exit()
		elif not converged:
			print('Error in def Place_Already_Created_Clusters_In_Population of class Population, in Population.py.')
			print("This Cluster did not converge when optimised. Check out why cluster "+str(cluster_name)+' did not converge by at the following output from the stdout.')
			print('*******************************************************')
			print(opt_information['output.txt'])
			print('*******************************************************')
			print('The program will also show the optimised and the unoptimised versions of cluster '+str(cluster_name))
			from ase.visualize import view; view(UnOpt_Cluster); view(Opt_Cluster)
			print('This program will end without doing anything.')
			exit()
		elif memory_operator.is_similar_cluster_in_memory_operator(Opt_Cluster):
			print('Error in def Place_Already_Created_Clusters_In_Population, in Initialise_Population.py.')
			print("Cluster "+str(cluster_name)+' in the population is too similar to one of the clusters in your memory operator database.')
			print('Check the clusters that you have included in your population and in the Memory Operator database')
			print('The program will now show the optimised version of cluster '+str(cluster_name))
			from ase.visualize import view; view(Opt_Cluster)
			print('This program will end without doing anything.')
			exit()	
		initialised_cluster = Cluster(Opt_Cluster)
		initialised_cluster.verify_cluster(cluster_name,0,vacuum_to_add_length,rounding_criteria)
		initialised_cluster.remove_calculator()
		# Set some of the details about the cluster that are needed for the population collection to save data to the population database.
		initialised_cluster.ever_in_population = True
		initialised_cluster.excluded_because_violates_predation_operator = False
		initialised_cluster.initial_population = True
		initialised_cluster.removed_by_memory_operator = False
		# Make sure this imported cluster has the correct chemical makeup for this genetic algorithm run.
		if not initialised_cluster.get_elemental_makeup() == cluster_makeup:
			print('Error in def Place_Already_Created_Clusters_In_Population of class Population, in Population.py.')
			print('Cluster '+str(cluster_name)+' does not contain the correct chemical makeup.')
			print('Cluster '+str(cluster_name)+' makeup: '+str(initialised_cluster.get_elemental_makeup()))
			print('Desired makeup: '+str(cluster_makeup))
			print('Check this out.')
			print('This program will end without doing anything')
			exit()
		index = cluster_name - 1
		print('Adding Cluster '+str(cluster_name)+' that the user wants to placed into the population of the GA, before the generation cycles of the GA begins.')
		population.add(index, initialised_cluster) # Place the initialised cluster into the population ## self.Add_ClusterToPopulation(initialised_cluster,index,UnOpt_Cluster=UnOpt_Cluster,Optimisation_Information=opt_information) 
	print('*****************************')
	##############################################################################################################################################
	# Check to make sure that the clusters that have been entered in by the user do not violate the diversity scheme.
	#clusters_removed_from_pop, removed_clusters_report = Fitness_Factor_Scheme.Check_Initial_Population(return_report=True)
	fitness_operator.assign_initial_population_fitnesses()
	clusters_to_create, removed_clusters_report = Check_Population_against_predation_operator(population,predation_operator)
	#import pdb; pdb.set_trace()
	if not len(clusters_to_create) == 0: #len(clusters_removed_from_pop):
		print('Error in def Place_Already_Created_Clusters_In_Population of class Population, in Population.py.')
		print('Some of the clusters that you have placed in the population folder violate the Diversity operator: '+str(predation_operator.Predation_Switch))
		print('Here is a list of the clusters that are the same as other clusters in this population:')
		for cluster_kept_dir, clusters_removed_dirs in sorted(removed_clusters_report,key=lambda x:x[0]):
			clusters_removed_dirs.insert(0,cluster_kept_dir)
			print('Clusters: '+str(clusters_removed_dirs)+' are considered identical by this diversity operator.')
		print('Check this out.')
		print('This program will end without doing anything')
		exit()
	##############################################################################################################################################
	print('Finished importing clusters from '+str(population.user_initialised_population_folder)+' into the population')
	return population[-1].name
	##############################################################################################################################################

# ----------------------------------------------------------------------------------------------------------------------------- #
# ----------------------------------------------------------------------------------------------------------------------------- #
# ----------------------------------------------------------------------------------------------------------------------------- #

def get_tasks(population,clusters_to_create,cell_length,vacuum_to_add_length,cluster_makeup,surface,r_ij,rounding_criteria,Minimisation_Function,memory_operator):
	"""
	This is a generator that will allow python to create clusters with multiprocessing.

	:param population: The population
	:type  population: Organisms.GA.Population
	:param clusters_to_create: This is a list of all the names for the clusters to be given.
	:type  clusters_to_create: list of int
	:param cell_length: This is the length of the square unit cell the cluster will be created in.
	:type  cell_length: float
	:param vacuum_to_add_length: The amount of vacuum to place around the cluster
	:type  vacuum_to_add_length: float
	:param cluster_makeup: The makeup of the cluster
	:type  cluster_makeup: dict.
	:param surface: This is the surface that the cluster is placed on. None means there is no surface.
	:type  surface: Organisms.GA.Surface
	:param r_ij: The maximum distance that should be between atoms to be considered bonded. This value should be as large a possible, to reflected the longest bond possible between atoms in the cluster.
	:type  r_ij: float
	:param rounding_criteria: The number of decimal places to round the energy of clusters to.
	:type  rounding_criteria: int
	:param Minimisation_Function: The minimisation function
	:type  Minimisation_Function: __func__
	:param memory_operator: The memory operator
	:type  memory_operator: Organisms.GA.Memory_Operator

	returns list of all the information needed to create a cluster
	rtype   list of Any
	"""
	def tasks(population_name, clusters_to_create, cell_length, vacuum_to_add_length, cluster_makeup, surface, r_ij, rounding_criteria, Minimisation_Function, memory_operator):
		for cluster_to_create in clusters_to_create:
			yield (population_name,cluster_to_create,cell_length,vacuum_to_add_length,cluster_makeup,surface,r_ij,rounding_criteria,Minimisation_Function,memory_operator)
	return tasks(population.name, clusters_to_create,cell_length,vacuum_to_add_length,cluster_makeup,surface,r_ij,rounding_criteria,Minimisation_Function,memory_operator)

import sys
from io import StringIO
from Organisms.GA.Types_Of_Mutations import randomMutate
from Organisms.GA.ExternalDefinitions import Exploded
def create_a_cluster(input_data):
	"""
	This will create a cluster, allowing the user to create cluster via multiprocessing.

	:param input_data: This tuple contains all the information required to create a new randomly generated cluster. 
	:type  input_data: list of Any

	returns tuple of the index to place the cluster into the population, and the optimised cluster
	rtype   (int,Organisms.GA.Cluster)
	"""
	(population_name,cluster_to_create,cell_length,vacuum_to_add_length,cluster_makeup,Surface,r_ij,rounding_criteria,Minimisation_Function,memory_operator) = input_data

	index, cluster_name = cluster_to_create
	#print("Creating Cluster " + str(cluster_name))
	# Make a folder to hold all information about the currently created cluster and move into that folder
	while True: # While statement required for the self.Diversity_switch == "CNA_Full" option and for the Exploded(Opt_Cluster) method
		try:
			UnOpt_Cluster = randomMutate(cell_length,vacuum_to_add_length,cluster_makeup=cluster_makeup)
			stdout = sys.stdout; output = StringIO(); sys.stdout = output
			Opt_Cluster, converged, opt_information = Minimisation_Function(UnOpt_Cluster.deepcopy(),population_name,cluster_name)
			#surface = Surface.get_surface(UnOpt_Cluster)
			#exit('bar1')
			sys.stdout = stdout
			opt_information['output.txt'] = output.getvalue()
		except Exception as exception:
			print('------------------------------------------')
			print('Error during def create_a_cluster in Population.py')
			print(str(exception)+'\n'+'Creating a new cluster.')
			print(str(exception)+'\n'+'Creating a new cluster.', file=sys.stderr)
			print('------------------------------------------')
			continue
		if Exploded(Opt_Cluster,max_distance_between_atoms=r_ij): # make sure the randomised cluster has not split up when optimised.
			print("Cluster exploded. Will obtain a new cluster " + str(cluster_name))
			continue # Since the cluster has exploded, we throw it out and repeat the process by making a new randomised cluster 
		elif not converged:
			print("Cluster did not converge. Will obtain a new cluster " + str(cluster_name))
			continue # Since the cluster did not converge, the cluster is not in a local minimum. This can be a problem its not what we want, and two atoms can be too close to eachother. 
		elif memory_operator.is_similar_cluster_in_memory_operator(Opt_Cluster):
			print("Cluster is too similar to a cluster in Memory Operator. Will obtain a new cluster " + str(cluster_name))
			continue
		else:
			break # break out of this cycle of for loop as we have found a cluster that satifsies the Explosion criteria and self.Diversity_switch is not == to "CNA_Full"

	Opt_Cluster.verify_cluster(cluster_name,0,vacuum_to_add_length,rounding_criteria)
	Opt_Cluster.remove_calculator()
	#print("Finished Creating Cluster " + str(cluster_name))
	return (index,Opt_Cluster)

# ----------------------------------------------------------------------------------------------------------------------------- #
# ----------------------------------------------------------------------------------------------------------------------------- #
# ----------------------------------------------------------------------------------------------------------------------------- #

import multiprocessing as mp
def Initialise_Population_with_Randomly_Generated_Clusters(population,cluster_makeup,surface,Minimisation_Function,cell_length,vacuum_to_add_length,r_ij,rounding_criteria,no_of_cpus,memory_operator,predation_operator,fitness_operator,previous_cluster_name):
	"""
	This method will place a number of randomly generated clusters into the population until it is at the desired size.

	:param population: The population
	:type  population: Organisms.GA.Population
	:param cluster_makeup: The makeup of the cluster
	:type  cluster_makeup: dict.
	:param surface: This is the surface that the cluster is placed on. None means there is no surface.
	:type  surface: Organisms.GA.Surface
	:param Minimisation_Function: The minimisation function
	:type  Minimisation_Function: __func__
	:param cell_length: This is the length of the square unit cell the cluster will be created in.
	:type  cell_length: float
	:param vacuum_to_add_length: The amount of vacuum to place around the cluster
	:type  vacuum_to_add_length: float
	:param r_ij: The maximum distance that should be between atoms to be considered bonded. This value should be as large a possible, to reflected the longest bond possible between atoms in the cluster.
	:type  r_ij: float
	:param rounding_criteria: The number of decimal places to round the energy of clusters to.
	:type  rounding_criteria: int
	:param no_of_cpus: The number of cpus available to create clusters 
	:type  no_of_cpus: int
	:param predation_operator: This is the predation operator 
	:type  predation_operator: Organisms.GA.Predation_Operator
	:param fitness_operator: This is the fitness operator
	:type  fitness_operator: Organisms.GA.Fitness_Operator
	:param memory_operator: The memory operator
	:type  memory_operator: Organisms.GA.Memory_Operator
	:param previous_cluster_name: This is the name of the last cluster created in the genetic algorithm.
	:type  previous_cluster_name: int

	returns The name of the most recently created cluster by this method in the population.
	rtype   int

	"""
	##############################################################################################################################################
	# get the dirs of the clusters that need to be populated.
	clusters_to_create = list(zip(range(population.size),range(previous_cluster_name+1,previous_cluster_name+population.size+1)))
	# This part is to prevent the population making new clusters that override the clusters that the user placed into the original population.
	current_cluster_names = population.get_cluster_names() # The current Dir tag of the cluster being inputted or created in the GA. Initially use to record all cluster Dir in run folder. Will be turned into an int once all original clusters in the population have been recorded into the GA program.
	for index in range(len(clusters_to_create)-1,-1,-1):
		if clusters_to_create[index][1] in current_cluster_names:
			del clusters_to_create[index]
	########################################################################################################
	# Initalise the rest of the population by creating randomised clusters using the InitaliseCluster method
	# from InitaliseCluster.py. The name of the dir tags will be those in the clusters_to_create list.
	while True:
		# This will create the required number of clusters to complete the population
		tasks = get_tasks(population,clusters_to_create,cell_length,vacuum_to_add_length,cluster_makeup,surface,r_ij,rounding_criteria,Minimisation_Function,memory_operator)
		'''
		with mp.Pool(processes=no_of_cpus) as pool: # pool = mp.Pool()
			results = pool.map_async(self.create_a_cluster, tasks)
			results.wait()
		made_clusters = results.get()
		'''
		# If you want to do work in serial rather than parallel
		made_clusters = []
		for task in tasks:
			try:
				made_cluster = create_a_cluster(task)
				made_clusters.append(made_cluster)
			except Exception as exception:
				print('Error in def Initialise_Population_with_Randomly_Generated_Clusters, in Initialise_Population.py')
				print('There was an issue with initialising the population with randomly generated clusters.')
				print('The issue was with the following cluster')
				print('\t --> '+str(task))
				print('Check this out, see the following exception for details')
				Lock_Remove()
				raise exception

		for made_cluster in made_clusters:
			index, Opt_Cluster = made_cluster
			Opt_Cluster.ever_in_population = True
			Opt_Cluster.excluded_because_violates_predation_operator = False
			Opt_Cluster.initial_population = True
			Opt_Cluster.removed_by_memory_operator = False
			population.add(index,Opt_Cluster) # Place the initialised cluster into the population
		#except:
		#	print('Weird error happening here? 2')
		#	import pdb; pdb.set_trace()
		#	exit()
		print("-----------------------------")
		# Assign fitnesses to clusters.
		fitness_operator.assign_initial_population_fitnesses()
		print('Checking that the clusters in the population satisfies the Diversity Scheme')
		# Check that the population does not violate the diversity scheme
		clusters_to_create, removed_clusters_report = Check_Population_against_predation_operator(population,predation_operator)
		if clusters_to_create == []:
			break
	return population[-1].name

# ----------------------------------------------------------------------------------------------------------------------------- #
# ----------------------------------------------------------------------------------------------------------------------------- #
# ----------------------------------------------------------------------------------------------------------------------------- #
# Initialise the population

def Initialise_Population(population,cluster_makeup,surface,Minimisation_Function,memory_operator,predation_operator,fitness_operator,epoch,cell_length,vacuum_to_add_length,r_ij,rounding_criteria,no_of_cpus,previous_cluster_name=0,generation=0,get_already_created_clusters=True,is_epoch=False,epoch_due_to_population_energy_convergence=None):
	"""
	This method will initalise the Population by
		1. Placing clusters that the user would like in the initial population
		2. Generate a number of extra clusters so that the population has the desired number of clusters in it.

	This method is used when the population is first created when the genetic algorithm is just starting, and when an epoch method is resetting the population with a new population of randomly generated clusters.

	:param population: The population
	:type  population: Organisms.GA.Population
	:param cluster_makeup: The makeup of the cluster
	:type  cluster_makeup: dict.
	:param surface: This is the surface that the cluster is placed on. None means there is no surface.
	:type  surface: Organisms.GA.Surface
	:param Minimisation_Function: The minimisation function
	:type  Minimisation_Function: __func__	
	:param memory_operator: The memory operator
	:type  memory_operator: Organisms.GA.Memory_Operator
	:param epoch: The epoch method
	:type  epoch: Organisms.GA.Epoch	
	:param predation_operator: This is the predation operator 
	:type  predation_operator: Organisms.GA.Predation_Operator
	:param fitness_operator: This is the fitness operator
	:type  fitness_operator: Organisms.GA.Fitness_Operator
	:param cell_length: This is the length of the square unit cell the cluster will be created in.
	:type  cell_length: float
	:param vacuum_to_add_length: The amount of vacuum to place around the cluster
	:type  vacuum_to_add_length: float
	:param r_ij: The maximum distance that should be between atoms to be considered bonded. This value should be as large a possible, to reflected the longest bond possible between atoms in the cluster.
	:type  r_ij: float
	:param rounding_criteria: The number of decimal places to round the energy of clusters to.
	:type  rounding_criteria: int
	:param no_of_cpus: The number of cpus available to create clusters 
	:type  no_of_cpus: int
	:param previous_cluster_name: This is the name of the last cluster created in the genetic algorithm.  Default: 0
	:type  previous_cluster_name: int
	:param generation: The number of generations that have been performed. Default: 0
	:type  generation: int
	:param get_already_created_clusters: Are there clusters that the user created in the population. True if yes, False if no. Default: True
	:type  get_already_created_clusters: bool.
	:param is_epoch: Has the genetic algorithm just epoched. Default: False
	:type  is_epoch: bool.
	:param epoch_due_to_population_energy_convergence: Did the genetic algorithm epochbecause the energies of the clusters in the last populatino converge. Default: None
	:type  epoch_due_to_population_energy_convergence: bool.

	:returns previous_cluster_name: This is the name of the last cluster created by this method
	:rtype   previous_cluster_name: int

	"""
	print('################################################################################################################')
	print('################################################################################################################')
	print("Initalising a Starting Population of Clusters if required to take pool up to a size of " + str(population.size))
	if get_already_created_clusters:
		print("Pool size to begin = " + str(population.get_pool_folder_size(population.user_initialised_population_folder)))
		previous_cluster_name = Place_Already_Created_Clusters_In_Population(population,cluster_makeup,Minimisation_Function,vacuum_to_add_length,r_ij,rounding_criteria,surface,memory_operator,predation_operator,fitness_operator,previous_cluster_name)
	print('################################################################################################################')
	if len(population) < population.size:
		print('Will now population the next '+str(population.size - len(population))+' with randomly generated clusters.')
		previous_cluster_name = Initialise_Population_with_Randomly_Generated_Clusters(population,cluster_makeup,surface,Minimisation_Function,cell_length,vacuum_to_add_length,r_ij,rounding_criteria,no_of_cpus,memory_operator,predation_operator,fitness_operator,previous_cluster_name)
		print('Have finished populating the rest of the population with '+str(population.size - len(population))+' randomly generated clusters.')
	# setting up the epoch and making sure initial population setup is all good
	if not is_epoch:
		if epoch.should_epoch(population,0):
			print('Error in def Initialise_Population in Initialise_Population.py')
			print('Your initial population without any generations somehow has epoched? Not sure how? Check this out')
			exit('Program will finish without completing')
	print('The population has been successfully initialised.')
	print('################################################################################################################')
	print('################################################################################################################')
	population.add_to_history_file(generation,is_epoch=is_epoch,epoch_due_to_population_energy_convergence=epoch_due_to_population_energy_convergence)
	population.current_state_file(generation)
	return previous_cluster_name



'''
Geoffrey Weal, Create_submitSL_slurm_Main.py, 10/02/2021

This program is designed to create the various forms of submit.sl/mass_submit.sl files that could be used to submit genetic algorithm trials to slurm.
'''

import os

def make_submitSL(local_path,project,time,nodes,ntasks_per_node,mem,partition='large',email='',python_version='Python/3.6.3-gimkl-2017a'):
    # create name for job
    print("creating submit.sl for "+str(local_path))
    name = local_path.replace('/','_')
    # writing the submit.sl script
    with open("submit.sl", "w") as submitSL:
        submitSL.write('#!/bin/bash -e\n')
        submitSL.write('#SBATCH -J ' + str(name) + '\n')
        submitSL.write('#SBATCH -A ' + str(project) + '         # Project Account\n')
        submitSL.write('#SBATCH --partition ' + str(partition) + '\n')
        submitSL.write('\n')
        submitSL.write('#SBATCH --time=' + str(time) + '     # Walltime\n')
        submitSL.write('#SBATCH --nodes=' + str(nodes) + '\n')
        submitSL.write('#On VASP, Ben Roberts recommends using the same number\n')
        submitSL.write('#of tasks on all nodes, even if this makes scheduling\n')
        submitSL.write('#a little more difficult\n')
        submitSL.write('#SBATCH --ntasks-per-node=' + str(ntasks_per_node) + '\n')
        submitSL.write('#SBATCH --mem=' + str(mem) + '\n')
        #submitSL.write('#SBATCH -C sb\n')
        submitSL.write('\n')
        #submitSL.write("#SBATCH --hint=nomultithread    # don't use hyperthreading"+'\n')
        submitSL.write('#SBATCH --output=slurm-%j.out      # %x and %j are replaced by job name and ID'+'\n')
        submitSL.write('#SBATCH --error=slurm-%j.err'+'\n')
        if not email == '':
            submitSL.write('#SBATCH --mail-user=' + str(email) + '\n')
            submitSL.write('#SBATCH --mail-type=ALL\n')
        submitSL.write('\n')
        submitSL.write('module load '+str(python_version)+'\n')
        submitSL.write('python Run.py\n')

def make_mass_submitSL_full(local_path,project,no_of_trials,time,nodes,ntasks_per_node,mem,partition='large',email='',RunMinimisation_Script_Name='RunMinimisation',python_version='Python/3.6.3-gimkl-2017a'):
    # create name for job
    print("creating mass_submit.sl for "+str(local_path))
    name = local_path.replace('/','_')
    # writing the mass_submit.sl script
    with open(local_path+'/'+"mass_submit.sl", "w") as submitSL:
        submitSL.write('#!/bin/bash -e\n')
        submitSL.write('#SBATCH -J ' + str(name) + '\n')
        submitSL.write('#SBATCH -A ' + str(project) + '         # Project Account\n')
        submitSL.write('\n')
        submitSL.write('#SBATCH --array=1-'+str(no_of_trials)+'\n')
        submitSL.write('\n')
        submitSL.write('#SBATCH --time=' + str(time) + '     # Walltime\n')
        submitSL.write('#SBATCH --nodes=' + str(nodes) + '\n')
        submitSL.write('#SBATCH --ntasks-per-node=' + str(ntasks_per_node) + '\n')
        submitSL.write('#SBATCH --mem=' + str(mem) + '\n')
        #submitSL.write('#SBATCH -C sb\n')
        submitSL.write('\n')
        submitSL.write('#SBATCH --partition='+str(partition)+'\n')
        submitSL.write('#SBATCH --output=arrayJob_%A_%a.out'+'\n')
        submitSL.write('#SBATCH --error=arrayJob_%A_%a.err'+'\n')
        if not email == '':
            submitSL.write('#SBATCH --mail-user=' + str(email) + '\n')
            submitSL.write('#SBATCH --mail-type=ALL\n')
        #submitSL.write('\n')
        #submitSL.write('#SBATCH --hint=nomultithread\n')
        submitSL.write('\n')
        submitSL.write('######################\n')
        submitSL.write('# Begin work section #\n')
        submitSL.write('######################\n')
        submitSL.write('\n')
        submitSL.write("# Print this sub-job's task ID\n")
        submitSL.write('echo "My SLURM_ARRAY_JOB_ID: "${SLURM_ARRAY_JOB_ID}\n')
        submitSL.write('echo "My SLURM_ARRAY_TASK_ID: "${SLURM_ARRAY_TASK_ID}\n')
        submitSL.write('\n')
        submitSL.write('module load '+str(python_version)+'\n')
        submitSL.write('\n')
        submitSL.write('if [ ! -d Trial${SLURM_ARRAY_TASK_ID} ]; then\n')
        submitSL.write('    mkdir Trial${SLURM_ARRAY_TASK_ID}\n')
        submitSL.write('fi\n')
        submitSL.write('cp Run.py Trial${SLURM_ARRAY_TASK_ID}\n')
        submitSL.write('cp '+str(RunMinimisation_Script_Name)+'.py Trial${SLURM_ARRAY_TASK_ID}\n')
        submitSL.write('cd Trial${SLURM_ARRAY_TASK_ID}\n')
        submitSL.write('python Run.py\n')
        submitSL.write('cd ..\n')
        submitSL.write('mv arrayJob_${SLURM_ARRAY_JOB_ID}_${SLURM_ARRAY_TASK_ID}.out Trial${SLURM_ARRAY_TASK_ID}\n')
        submitSL.write('mv arrayJob_${SLURM_ARRAY_JOB_ID}_${SLURM_ARRAY_TASK_ID}.err Trial${SLURM_ARRAY_TASK_ID}\n')
        submitSL.close()

def make_mass_submitSL_packets(local_path,project,no_of_trials,no_of_packets_to_make,time,nodes,ntasks_per_node,mem,partition='large',email='',RunMinimisation_Script_Name='RunMinimisation',python_version='Python/3.6.3-gimkl-2017a'):
    # make sure that no_of_packets_to_make is a value and no_of_trials is divisible by it. 
    if not isinstance(no_of_packets_to_make,int):
        print('Error in def make_mass_submitSL_packets, in Create_submitSL_slurm_Main.py')
        print('no_of_packets_to_make needs to be give as a int')
        print('no_of_packets_to_make = '+str(no_of_packets_to_make))
        exit('Check this. The algorithm will finish here.')
    if not (float(no_of_trials)%float(no_of_packets_to_make)) == 0.0:
        print('Error in def make_mass_submitSL_packets, in Create_submitSL_slurm_Main.py')
        print('no_of_trials needs to be divisible by no_of_packets_to_make')
        print('no_of_trials = '+str(no_of_trials))
        print('no_of_packets_to_make = '+str(no_of_packets_to_make))
        exit('Check this. The algorithm will finish here.')
    number_of_divides = int(float(no_of_trials)/float(no_of_packets_to_make))
    # create name for job
    print("creating mass_submit.sl for "+str(local_path))
    name = local_path.replace('/','_')
    # writing the mass_submit.sl script
    with open(local_path+'/'+"mass_submit.sl", "w") as submitSL:
        submitSL.write('#!/bin/bash -e\n')
        submitSL.write('#SBATCH -J ' + str(name) + '\n')
        submitSL.write('#SBATCH -A ' + str(project) + '         # Project Account\n')
        submitSL.write('\n')
        submitSL.write('#SBATCH --array=1-'+str(no_of_packets_to_make)+'\n')
        submitSL.write('\n')
        submitSL.write('#SBATCH --time=' + str(time) + '     # Walltime\n')
        submitSL.write('#SBATCH --nodes=' + str(nodes) + '\n')
        submitSL.write('#SBATCH --ntasks-per-node=' + str(ntasks_per_node) + '\n')
        submitSL.write('#SBATCH --mem=' + str(mem) + '\n')
        #submitSL.write('#SBATCH -C sb\n')
        submitSL.write('\n')
        submitSL.write('#SBATCH --partition='+str(partition)+'\n')
        submitSL.write('#SBATCH --output=arrayJob_%A_%a.out'+'\n')
        submitSL.write('#SBATCH --error=arrayJob_%A_%a.err'+'\n')
        if not email == '':
            submitSL.write('#SBATCH --mail-user=' + str(email) + '\n')
            submitSL.write('#SBATCH --mail-type=ALL\n')
        #submitSL.write('\n')
        #submitSL.write('#SBATCH --hint=nomultithread\n')
        submitSL.write('\n')
        submitSL.write('######################\n')
        submitSL.write('# Begin work section #\n')
        submitSL.write('######################\n')
        submitSL.write('\n')
        submitSL.write("# Print this sub-job's task ID\n")
        submitSL.write('echo "My SLURM_ARRAY_JOB_ID: "${SLURM_ARRAY_JOB_ID}\n')
        submitSL.write('echo "My SLURM_ARRAY_TASK_ID: "${SLURM_ARRAY_TASK_ID}\n')
        submitSL.write('\n')
        submitSL.write('module load '+str(python_version)+'\n')
        submitSL.write('\n')
        submitSL.write('number_of_divides='+str(number_of_divides)+'\n')
        submitSL.write('for i in $( eval echo {1..${number_of_divides}} ); do\n')
        submitSL.write('\n')
        submitSL.write('trial_no=$(( $(( $(( ${SLURM_ARRAY_TASK_ID} - 1)) * ${number_of_divides} )) + $i ))\n')
        submitSL.write('echo Currently performing caluclation on trial: $trial_no\n')
        submitSL.write('\n')
        submitSL.write('if [ ! -d Trial${trial_no} ]; then\n')
        submitSL.write('    mkdir Trial${trial_no}\n')
        submitSL.write('fi\n')
        submitSL.write("echo '=============================================================================='\n")
        submitSL.write("echo '=============================================================================='\n")
        submitSL.write("echo 'Running Trial${trial_no}\n")
        submitSL.write("echo '=============================================================================='\n")
        submitSL.write("echo '=============================================================================='\n")
        submitSL.write("1>&2 echo '=============================================================================='\n")
        submitSL.write("1>&2 echo '=============================================================================='\n")
        submitSL.write("1>&2 echo 'Running Trial${trial_no}\n")
        submitSL.write("1>&2 echo '=============================================================================='\n")
        submitSL.write("1>&2 echo '=============================================================================='\n")
        submitSL.write('cp Run.py Trial${trial_no}\n')
        submitSL.write('cp '+str(RunMinimisation_Script_Name)+'.py Trial${trial_no}\n')
        submitSL.write('cd Trial${trial_no}\n')
        submitSL.write('python Run.py\n')
        submitSL.write("echo '=============================================================================='\n")
        submitSL.write("1>&2 echo '=============================================================================='\n")
        submitSL.write('cd ..\n')
        submitSL.write('cp arrayJob_${SLURM_ARRAY_JOB_ID}_${SLURM_ARRAY_TASK_ID}.out Trial${trial_no}/arrayJob_${SLURM_ARRAY_JOB_ID}_${trial_no}.out\n')
        submitSL.write('cp arrayJob_${SLURM_ARRAY_JOB_ID}_${SLURM_ARRAY_TASK_ID}.err Trial${trial_no}/arrayJob_${SLURM_ARRAY_JOB_ID}_${trial_no}.err\n')
        submitSL.write('echo -n "" > arrayJob_${SLURM_ARRAY_JOB_ID}_${SLURM_ARRAY_TASK_ID}.out\n')
        submitSL.write('echo -n "" > arrayJob_${SLURM_ARRAY_JOB_ID}_${SLURM_ARRAY_TASK_ID}.err\n')
        submitSL.write('\n')
        submitSL.write('done\n')
        submitSL.close()


import copy
import re
from random import randrange, uniform, randint, random
from ase import Atoms

from Organisms.GA.Types_Of_Mutations import moveMutate, homotopMutate, randomMutate
from Organisms.GA.Population import Population
from Organisms.GA.Cluster import Cluster
from Organisms.GA.ExternalDefinitions import InclusionRadiusOfCluster

population_type_example = Population('example',-1,user_initialised_population_folder=None,write_data=False)
cluster_type_example = Cluster()
atoms_type_example = Atoms()

def isfloat(element):
	"""
	Is the input a float.

	:param element: some input
	:type  element: Any

	:returns True if the element can be considered a float, False if not. 
	:rtype   ebool.
	"""
	if re.match(r'^-?\d+(?:\.\d+)?$', element) is None:
		return False
	else:
		return True

class Mutation:
	"""
	This class contains all the information and proceedures required to perform a Mutation. 

	:param mutation_types: Contains the information about the mutations that the user would like to perform. See Manual for more information about this.
	:type  mutation_types: [(str, float),...]
	:param r_ij: The maximum distance that should be between atoms to be considered bonded. This value should be as large a possible, to reflected the longest bond possible between atoms in the cluster.
	:type  r_ij: float
	:param vacuum_to_add_length: The amount of vacuum to place around the cluster
	:type  vacuum_to_add_length: float
	"""
	def __init__(self,mutation_types,r_ij,vacuum_to_add_length):
		self.mutation_types = mutation_types
		#self.r_ij = r_ij
		#self.vacuum_to_add_length = vacuum_to_add_length
		self.check(r_ij)
		self.change_mutation_chances() # update the values in self.mutation_types into the format used by Mutation class.

	def check(self, r_ij):
		"""
		This method will check that the inputs for the mutation_types has been done correctly, without any violating issues.

		:param r_ij: The maximum distance that should be between atoms to be considered bonded. This value should be as large a possible, to reflected the longest bond possible between atoms in the cluster.
		:type  r_ij: float
		"""
		mutation_types, mutation_chances = [list(i) for i in zip(*self.mutation_types)]
		# check that all mutation types are accepted types.
		# DEAL WITH LATER
		for mutation_type in mutation_types:
			if mutation_type.startswith('move'):
				if mutation_type == 'move':
					self.dist_to_move = float(r_ij)*0.5
				elif mutation_type.startswith("move_") and isfloat(mutation_type.replace("move_",'')):
					self.dist_to_move = float(mutation_type.replace("move_",''))
				else:
					print('Error in class MutationProcedure, in MutationProcedure.py')
					print('You need to enter the move entry in either by:')
					print('    * "move", where the move distance is set to default (r_ij*0.5).')
					print('    * "move_XX", where XX is a float that indicate the maximum move distance.')
					print('You input for the move method: '+str(mutation_type))
					print('Check this out. This program will now exit.')
					exit('This program will finish without completing.')
			elif mutation_type.startswith('homotop'):
				if not len(self.population.cluster_makeup) > 1:
					print('Error in class MutationProcedure, in MutationProcedure.py')
					print('You want to use the mutation method "homotop", but your cluster is monometallic')
					print('Cluster Makeup: '+str(self.population.cluster_makeup))
					print('Check this out.')
					exit('This program will finish without completing.')
			elif mutation_type.startswith('random'):
				pass
				#if self.boxtoplaceinlength == None:
				#	print('Error in def mutation, in Class MutationProceedure, in MutationProceedure.py.')
				#	print('If you want to use the "random" mutation method, you need to specify a value for the boxtoplaceinlength variable.')
				#	print('Check this and see if you need to set a value for the boxtoplaceinlength variable.')
				#	exit('This program will finish without completing.')
			elif mutation_type.startswith("random_"):
				continue
			else:
				print('Error in class MutationProcedure, in MutationProcedure.py')
				print('mutation_type '+str(mutation_type)+' is not one of the types of mutation methods')
				print("The types of mutation_types available are: 'move','homotop','random'")
				print('Check this.')
				exit('This program will finish without completing.')
		# Make sure that there are no repetitions of the mutation methods in mutTypes.
		if not len(mutation_types) == len(set(mutation_types)):
			print('Error in class MutationProcedure, in MutationProcedure.py')
			print('You have entered a repeated mutType into mutTypes')
			print('mutTypes: '+str(self.mutTypes))
			print('Check this.')
			exit('This program will finish without completing.')
		# Check if the total chance of mutTypes is equal to 1.
		total_mutation_type_to_choose = sum(mutation_chances)
		if not total_mutation_type_to_choose == 1.0:
			print('Error: the total chance in mutType must equal 1.')
			print('Check your mutType variable')
			print('mutTypes = ' + str(mutTypes))
			print('Total mutTypes chance = ' + str(total_mutation_type_to_choose))
			import pdb; pdb.set_trace()
			exit('This program will finish without completing.')

	def change_mutation_chances(self):
		"""
		This method will change the format of self.mutation_types into a format that can be used by this Mutation class.

		The format of self.mutation_types changes as follows with this example.
			[] => []
		"""
		end_point = 0.0
		for mutation_type in self.mutation_types:
			end_point += mutation_type[1]
			end_point = round(end_point,12)
			mutation_type[1] = end_point
		if not self.mutation_types[-1][1] == 1.0:
			print('Error: The total chance of picking any mutation_type does not equal 1.')
			print('Check your mutation_type variable')
			print('self.mutation_types = ' + str(self.mutation_types))
			print('total  chance = ' + str(self.mutation_types[-1][1]))
			print('Check this')
			import pdb; pdb.set_trace()
			exit('This program will finish without completing.')

	def run(self,run_input,cell_length=None,vacuum_length=None):
		'''
		This definition will run the Mutation Proceedure. 

		Inputs:
			run_input (GA.Population/GA.Cluster/ase.Atoms): This is the input data about the cluster to mutate, or the population to take a cluster to mutate.
			cell_length (float): This is the cell length of the unit cell.
			vacuum_length (float): The amount of vacuum to add to the unit cell.

		Returns:
			mutant (GA.Cluster) - The mutated cluster
			mutation_method (str.) - The type of mutation that was picked to use for this mutation.
		'''
		if type(run_input) is type(population_type_example):
			cluster_to_mutate = self.pickClusterFromThePopulation(run_input)
		elif type(run_input) is type(cluster_type_example) or type(run_input) is type(atoms_type_example):
			cluster_to_mutate = run_input
		else:
			print('Error in def run of class Mutation.')
			print('run_input must be either a population or a specific cluster (of type ASE.Atoms or Cluster).')
			print('run_inpu: '+str(run_inpu))
			print('Check this')
			import pdb; pdb.set_trace()
			exit('This program will finish without completing.')
		#print('The cluster to be mutated is cluster: '+str(cluster_to_mutate.name))
		# Pick the type of mutation method to use
		mutation_method = self.get_mutation_type()
		# perform the mutation. 
		mutant = self.mutation(mutation_method,cluster_to_mutate,cell_length,vacuum_length)
		# respecify the vacuum around the cluster to that its distance is vacuumAdd
		lengthOfCell = 2.0*InclusionRadiusOfCluster(mutant) + vacuum_length
		cell = [lengthOfCell,lengthOfCell,lengthOfCell]
		mutant.set_cell(cell)
		mutant.center() # centre the cluster within this unit cell
		# return the mutant, as well as the mutation method that was performed. 
		if not len(mutant) == len(cluster_to_mutate):
			print('Error in def run, in Mutation.py')
			print('The offspring contains '+str(len(mutant))+' atoms, but should contain '+str(len(cluster_to_mutate)))
			print('Check this')
			import pdb; pdb.set_trace()
			print('This program will finish without completing')
			exit()
		return mutant, mutation_method

	def pickClusterFromThePopulation(self,population):
		"""
		This definition will pick the cluster from the population to mutate.

		Inputs:
			population (GA.Population): The population to choose the cluster to mutate from.
		
		:returns: cluster_to_mutate - The list of parents for mating.
		:rtypes: Cluster

		"""
		index = randrange(0,len(population))
		cluster_to_mutate = population[index].deepcopy()
		return cluster_to_mutate

	def get_mutation_type(self):
		"""
		This definition will pick the type of mutation.

		:returns: mutation_type
		:rtypes: str.

		"""
		if not len(self.mutation_types) == 1:
			chance_of_type_of_mutation = uniform(0,1)
		else:
			chance_of_type_of_mutation = 0 # dont need to perform uniform(0,1) if you only have one mutation_method
		for mutation_type, mutation_echelon in self.mutation_types:
			if chance_of_type_of_mutation <= mutation_echelon:
				return mutation_type
		print('Error in def get_mutation_type in class Mutation_Procedure, in MutationProcedure.py')
		print('Somehow, I think chance_of_type_of_mutation is greater than 1???')
		print('chance_of_type_of_mutation = '+str(chance_of_type_of_mutation))
		print('self.mutation_types = '+str(self.mutation_types))
		print('Check this')
		import pdb; pdb.set_trace()
		exit('This program will finish without completing.')

	def mutation(self,mutation_method,cluster_to_mutate,cell_length=None,vacuum_length=None):
		"""
		This method will perform the mutation and return the mutant. 

		Inputs:
			mutation_method (str.): The type of mutation to use for the mutation
			cluster_to_mutate (GA>Cluster): This is the cluster that we would like to perform a mutation upon.
			cell_length (float): This is the cell length of the unit cell.
			vacuum_length (float): The amount of vacuum to add to the unit cell.

		:returns: mutant
		:rtypes: Atom

		"""
		if mutation_method == "random":
			mutant = randomMutate(cell_length, vacuum_length,cluster_makeup=cluster_to_mutate.get_elemental_makeup(),cluster_to_mutate=None,percentage_of_cluster_to_randomise=None)
		elif mutation_method.startswith('random_'):
			percentage_of_cluster_to_randomise = mutation_method.replace('random_','')
			try:
				percentage_of_cluster_to_randomise = float(percentage_of_cluster_to_randomise)
			except:
				print('Error in MutationProcedure Class: If you are using chosen_mutation_type in random_X where X is a percentage of atoms to be randomised, X must be in the form of an interger or a decimal number.')
				print('Check this.')
				print('chosen_mutation_type: '+str(mutation_method))
				import pdb; pdb.set_trace()
				exit()
			# this should be check later to see if it is working as intended.
			mutant = randomMutate(cell_length, vacuum_length, cluster_makeup=None,cluster_to_mutate=cluster_to_mutate,percentage_of_cluster_to_randomise=percentage_of_cluster_to_randomise)
		elif mutation_method.startswith("move"):
			mutant = moveMutate(cluster_to_mutate,self.dist_to_move)
		elif mutation_method == "homotop":
			mutant = homotopMutate(cluster_to_mutate)
		else:
			print('Check this')
			import pdb; pdb.set_trace()
			exit('This program will finish without completing.')
		if not len(mutant) == len(cluster_to_mutate):
			print('Error in def mutation, in Mutation.py')
			print('The offspring contains '+str(len(mutant))+' atoms, but should contain '+str(len(cluster_to_mutate)))
			print('Check this')
			import pdb; pdb.set_trace()
			print('This program will finish without completing')
			exit()
		return mutant



from random import uniform 
import os, sys
from io import StringIO
from time import time

from Organisms.GA.ExternalDefinitions import Exploded

def Create_An_Unoptimised_Offspring(cluster_number,chance_of_mutation,cell_length,vacuum_to_add_length,population,creating_offspring_mode,crossover_procedure,mutation_procedure):
	"""
	This definition provides the methodology for how to create an offspring.

	Inputs:
		cluster_number (int): the name of the cluster
		chance_of_mutation (float): the change than a mutation will occur, either as well as or instead of crossover, depending on input for creating_offspring_mode.
		cell_length (float): The size of the unit cell.
		vacuum_to_add_length (float)" The amount of vacuum to add to the unit cell"
		population (Organisms.GA.Population): This is the population that clusters will be taken from for mating and mutation schemes.
		creating_offspring_mode (str.): This tag indicates how the offspring are created. 

			* creating_offspring_mode == 'Mating_and_Mutation_Together' - The mating scheme is run first, followed by a mutation scheme with some probability.. 
			* creating_offspring_mode == 'Either_Mating_and_Mutation' - Either the mating 'or' the mutation scheme will run, depending on a probability value.
		
		crossover_procedure (Organisms.GA.Crossover): 
		mutation_procedure (Organisms.GA.Mutation): 

	:returns: the newly created offspring cluster.
	:rtype: Organisms.GA.Cluster
	"""
	# Choose to perform a Mutation or Mating Proceedure
	choice_mate_or_mutate = uniform(0,1)
	if creating_offspring_mode == 'Mating_and_Mutation_Together':
		#print("Cluster " + str(cluster_number) + " will be put through a Mating Method.")
		offspring = crossover_procedure.run(population)
		#how_created.append('Mating')
		if Will_Mutation_Occur(choice_mate_or_mutate,chance_of_mutation):
			#print("Cluster " + str(cluster_number) + " will be put through a Mutation Method.")
			offspring, mutation_method = mutation_procedure.run(cluster_to_mutate=cluster_to_mutate,cell_length=cell_length,vacuum_length=vacuum_to_add_length)
			#how_created.append('Mutation')
			#print('The mutation method used was ' + mutation_method)
	elif creating_offspring_mode == 'Either_Mating_and_Mutation':
		if Will_Mutation_Occur(choice_mate_or_mutate,chance_of_mutation):
			#print("Cluster " + str(cluster_number) + " obtained using the Mutation Method.")
			offspring, mutation_method = mutation_procedure.run(population,cell_length=cell_length,vacuum_length=vacuum_to_add_length)
			#how_created.append('Mutation')
			#print('The mutation method used was ' + mutation_method)
		else:
			#print("Cluster " + str(cluster_number) + " obtained using the Mating Method.")
			offspring = crossover_procedure.run(population)
			#how_created.append('Mating')
	else:
		print('creating_offspring_mode must be either Mating_and_Mutation_Together or Either_Mating_and_Mutation')
		print('Check this')
		print('creating_offspring_mode = ' + str(creating_offspring_mode))
		exit('This program will finish without completion.')
	if not len(offspring) == len(population[0]):
		print('Error in def Create_An_Unoptimised_Offspring, in Get_Offspring.py')
		print('The offspring contains '+str(len(offspring))+' atoms, but should contain '+str(len(population[0])))
		print('Check this')
		import pdb; pdb.set_trace()
		print('This program will finish without completing')
		exit()
	return offspring

def Will_Mutation_Occur(choice_mate_or_mutate,chance_of_mutation):
	"""
	This method will determine if the algotithm should perform a mutation.

	:param choice_mate_or_mutate: This is a random value between 0 and 1. If this value is less than self.chance_of_mutation, this indicates to perform a mutation
	:type  choice_mate_or_mutate: float
	:param chance_of_mutation: This a decimal value between 0 and 1 that determines the chance of a mutation.
	:type  chance_of_mutation: float

	:returns: If True; mutate. If False; do not mutate (mate).
	:rtype: bool.
	"""
	return choice_mate_or_mutate < chance_of_mutation
	
def Create_An_Offspring(input_data):
	"""
	This method is used to obtain an offspring

	:param input_data: A tuple of all the information needed to make a new offspring
	:type  input_data: tuple of many inputs

	:returns: The offspring and a string containing any information that would be useful for the user to see about how this cluster was made
	:rtype: Organisms.GA.Cluster and str.
	"""
	#total_start_time = time()
	(run_number,generation_number,population,offspring_pool_name,chance_of_mutation,r_ij,cell_length,vacuum_to_add_length,creating_offspring_mode,crossover_procedure,mutation_procedure,no_offspring_per_generation,rounding_criteria,Minimisation_Function,surface,place_cluster_where,print_details) = input_data
	toString = ''
	cluster_name = run_number
	#if print_details:
	#	toString += "----------------------------------------------------------"+'\n'
	no_of_explosions = 0; no_of_not_converged = 0
	while True:
		#cluster_start_time = time()
		#stdout = sys.stdout; output = StringIO(); sys.stdout = output
		UnOpt_offspring = Create_An_Unoptimised_Offspring(cluster_name,chance_of_mutation,cell_length,vacuum_to_add_length,population,creating_offspring_mode,crossover_procedure,mutation_procedure)
		#sys.stdout = stdout
		#cluster_end_time = time()
		#if print_details:
		#	toString += output.getvalue().rstrip()
		#######################################################################################
		# Locally Minimise the cluster. Since the population class looks after the population folder, this process is done from self.population
		#minimisation_start_time = time()
		#stdout = sys.stdout; output = StringIO(); sys.stdout = output
		try_statement_counter = 0
		while True:
			try:
				Opt_offspring, converged, opt_information = Minimisation_Function(UnOpt_offspring.deepcopy(),offspring_pool_name,cluster_name)
				break
			except Exception as exception:
				print('Error during minimisation. Something has gone wrong during the local optimisation process',file=sys.stderr)
				print('Generation: '+str(generation_number)+'; Cluster Name: '+str(run_number),file=sys.stderr)
				print('exception',file=sys.stderr)
				print('Will Try to run the local optimisation process again',file=sys.stderr)
				try_statement_counter += 1
			if try_statement_counter == 10:
				print('The above error has occurred 10 times.',file=sys.stderr)
				print('Will make a new offspring and discard this one',file=sys.stderr)
				break
		#sys.stdout = stdout
		#if print_details:
		#	opt_information['output.txt'] = output.getvalue()
		#######################################################################################
		if not try_statement_counter == 10:
			#minimisation_end_time = time()
			break
			#exploded_start_time = time()
			did_explode = Exploded(Opt_offspring,max_distance_between_atoms=r_ij)
			#exploded_end_time = time()
			if did_explode: # make sure the randomised cluster has not split up when optimised.
				toString += "Cluster exploded. Will disregard this cluster and try mating or mutation method again"+'\n'
				no_of_explosions += 1
			elif not converged:
				toString += "Cluster did not converge. Will disregard this cluster and try mating or mutation method again"+'\n'
				no_of_not_converged += 1
			else:
				break
	# Add the cluster to the list of offspring from this list
	#extra1_start = time()
	Opt_offspring.verify_cluster(cluster_name,generation_number,vacuum_to_add_length,rounding_criteria)
	#extra1_end = time()
	#extra2_start = time()
	Opt_offspring.remove_calculator()
	'''
	extra2_end = time()
	minimisation_end_time = time()
	#print(str(cluster_name)+', ', end = '')
	total_end_time = time()
	print('Total offspring creation time: '+str(total_end_time-total_start_time)+' s.')
	print('Make offspring time: '+str(cluster_end_time-cluster_start_time)+' s. Run Minimisation time: '+str(minimisation_end_time-minimisation_start_time)+' s.')#' Check for exploded: '+str(exploded_end_time-exploded_start_time)+' s.')
	print('Extra time 1: '+str(extra1_end-extra1_start)+' s.')
	print('Extra time 2: '+str(extra2_end-extra2_start)+' s.')
	print('---------------------------------------------------')
	'''
	return Opt_offspring, toString

	


'''
No_Predation_Operator.py, Geoffrey Weal, 28/10/2018

This is one of the Predation Operators that can be used by the genetic algorithm program. 

'''

from Organisms.GA.Predation_Operators.Predation_Operator import Predation_Operator

class No_Predation_Operator(Predation_Operator):
	"""
	This is one of the Predation Operators that can be used by the genetic algorithm program. 

	This Predation Operator will not do anything. It is the option to pick if you do not want to remove offspring or clusters from the population due to being energetically or strucutrally similar.

	:param Predation_Information: This contains all the information needed by the Predation Operator you want to use to run.
	:type  Predation_Information: dict.
	:param population: This is the population that this Operator will be controlling to make sure that no two clusters in the population have the same energy.
	:type  population: Organisms.GA.Population
	:param print_details: Print details of the predation operator, like verbose
	:type  print_details: bool

	"""
	def __init__(self,Predation_Information,population,print_details):
		super().__init__(Predation_Information,population,print_details)
		if not Predation_Information['Predation Operator'] == 'Off':
			print('Error in class No_Predation_Operator, in No_Predation_Operator.py')
			print('The Predation Operator "No_Predation_Operator" has been initialised.')
			print("However, Predation_Information['Predation Operator'] is not 'Off''")
			print("Predation_Information['Predation Operator'] = "+str(Predation_Information['Predation_Operator']))
			print('Check this out.')
			import pdb; pdb.set_trace()
			exit()

	###########################################################################################################
	########################## Methods Required for the def check_initial_population ##########################
	###########################################################################################################

	def check_initial_population(self,return_report=False):
		"""
		This definition is responsible for making sure that the initialised population obeys the Energy Predation Operator. 

		Since this 'Operator' does nothing, this def will not do anything. It will report back that it has not removed any cluster

		:param return_report: Will return a dict with all the information about what clusters are similar to what other clusters in the population. 
		:type  return_report: bool.

		returns:
			* **clusters_to_remove** (*list of ints*): a list of the clusters to remove from the population as they violate the Predation Operator. Format is [(index in population, name fo cluster),...]
			* **CNA_report** (*dict.*): a dictionary with information on the clusters being removed and the other clusters in the population which have caused the violation to the SCM Predation Operator. This information is only used to display information so they know why there are violations to the Predation Operator when they occur. For is {removed cluster: [list of clusters that this cluster is similar to in the population.]}
		
		"""
		if return_report:
			return [], {}
		else:
			return []

	########################################################################################################
	########################## Methods Required for the def assess_for_violations ##########################
	########################################################################################################

	def assess_for_violations(self,offspring_pool,force_replace_pop_clusters_with_offspring):
		"""
		This definition is designed to determine which offspring (and the clusters in the population) violate the predation Operator. 
		It will not remove or change any clusters in the offspring or population, but instead will record which offspring violate the 
		predation Operator. 

		Since this 'Operator' does nothing, this def will not do anything. 

		It will return two tuples with nothing in them, as required by this def. 

		:param offspring_pool: This is the collection of offspring to assess for violations to the Predation Operator.
		:type  offspring_pool: Organisms.GA.Offspring_Pool.Offspring_Pool
		:param force_replace_pop_clusters_with_offspring: This will tell the genetic algorithm whether to swap clusters in the population with offspring if the predation operator indicates they are the same but the predation operator has a better fitness value than the cluster in the population. 
		:type  force_replace_pop_clusters_with_offspring: bool.

		returns:
			* **offspring_to_remove** (*tuple of ints*): A list of the names of the offspring to be removed
			* **force_replacement** (*tuple of (int, int)*): A list of the clusters in the population that should be replaced, and the offspring they should be replaced by.

		"""
		return [], []

	########################################################################################################
	########################################################################################################
	########################################################################################################

	def add_to_database(self, collection):
		'''
		Add clusters similarities to the CNA database to be stored for future generations. 

		This does not do anything since this predation method does notthing

		:param collection: update the fitnesses of clusters in the collection.
		:type  collection: Organisms.GA.Collection.Collection
		'''
		pass

	def remove_from_database(self, cluster_names_to_remove):
		'''
		Clusters to remove from the CNA database

		This does not do anything since this predation method does notthing

		:param cluster_names_to_remove: A list of the names of all the clusters to remove from the CNA database
		:type  cluster_names_to_remove: list of ints
		'''
		pass

	def reset(self):
		'''
		Reset the CNA database with no inputs

		This does not do anything since this predation method does notthing
		'''
		pass



'''
CNA_Entry.py, Geoffrey Weal, 29/10/2018

This class holds a CNA comparison entry between two clusters, Cluster cluster_1 and Cluster cluster_2, including
the similarity_profile for different values of rCut, the average and the result of the comparison, 
based on the result from Geoff's Thesis

'''
import sys

def get_comparison(percentage_similarities):
	'''
	From the information from self.percentage_similarities, this method will write the result of the 
	comparison into self.similarity_category, as determined from the proceedure created by Geoffrey Weal.

	:param percentage_similarities: These are a list of all the similarities from performing the SCM across many rCut values.
	:type  percentage_similarities: list of floats

	:returns: The similarity class that the pair of clusters is assigned to based on the percentage_similarities, and the maximum similarity of percentage_similarities
	:rtype:   a str. and a float

	'''
	maximum_similarity = max(percentage_similarities)
	if   maximum_similarity == 100.0:
		similarity_category = 'geometric'
	elif maximum_similarity >= 50.0:
		similarity_category = 'same motif'
	else:
		similarity_category = 'different'
	return similarity_category, maximum_similarity

class Similarity_Profile:
	'''
	This class holds a CNA comparison entry between two clusters, Cluster 1 and Cluster 2, including
	the similarity_profile for different values of rCut, the average and the result of the comparison, 
	based on the result from Geoff's Thesis

	:param name_1: Name of the first cluster.
	:type  name_1: int
	:param name_2: Name of the first cluster.
	:type  name_2: int
	:param similarity_profile: The similarity profile between cluster 1 and cluster 2.
	:type  similarity_profile: list of floats
	
	'''
	def __init__(self,name_1, name_2, similarity_profile):
		if name_1 < name_2:
			self.name_1 = name_1; self.name_2 = name_2; 
		elif name_1 > name_2:
			self.name_1 = name_2; self.name_2 = name_1; 
		else:
			exit('Error')
		self.similarity_profile = similarity_profile
		self.get_results()

	def __repr__(self):
		"""
		This is what the terminal will print to represent your class. 
		"""
		toString = str(self.name_1)+'Vs.'+str(self.name_2)
		return toString

	def get_results(self):
		'''
		Get the result of the average of similarity_profile, and the result of the comparison.
		'''
		self.get_comparison()
		self.get_average()

	def get_average(self):
		'''
		Get the average of similarity_profile
		'''
		self.average = float(sum(self.similarity_profile))/float(len(self.similarity_profile))

	def get_comparison(self):
		'''
		From the information from self.similarity_profile, this method will write the result of the 
		comparison into self.similarity_category, as determined from the proceedure created by Geoffrey Weal.
		'''
		self.similarity_category, self.similarity_max = get_comparison(self.similarity_profile)












def get_average(fitnesses):
	sum_of_fitnesses = sum(fitnesses)
	average = float(sum_of_fitnesses)/len(fitnesses)
	return average

def get_population_fitness(cluster_SCM_simiarities,population_fitness_function):
	population_average_similarity = get_average(cluster_SCM_simiarities)
	population_fitness = population_fitness_function.get_fitness(population_average_similarity)
	return population_fitness





class Collections_Iterator:
	"""
	This iterator is designed to iterate through mutliple collections together, in such a way that we sample every cluster across a range of collections.
	This method is designed to work in place so that it take up minimal amount of space on ram.

	:param population: This is the population being used for this Genetic Run
	:type  population: Organisms.GA.Population
	:param offspring_pools: These are all the offspring_pools that are being used. This can be inputs as one offspring_pool, or a list of multiple offspring_pools
	:type  offspring_pools: Organisms.GA.Offspring_Pool or [Organisms.GA.Offspring_Pool,...]
	:param pass_first_cluster: Start with the first cluster in the population, or the second.
	:type  pass_first_cluster: bool.

	"""
	def __init__(self, population, offspring_pools=None, pass_first_cluster=False):
		self.population = population
		self.collections = [population]

		if not offspring_pools == None:
			if not isinstance(offspring_pools,list):
				self.collections += [offspring_pools]
			else:
				self.collections += offspring_pools

		self.length = len(self.population)
		if not offspring_pools == None:
			for offspring_pool in offspring_pools:
				self.length += len(offspring_pool)

		if pass_first_cluster == False:
			self.index = 0
			self.index_collections = 0
		else:
			self.index = 1
			if self.index >= len(self.collections[0]):
				self.index_collections = 1
			else:
				self.index_collections = 0

	def __iter__(self):
		return self

	def __next__(self):
		"""
		Move through the collections to return the next cluster from all the collections.

		Return
			cluster(Organisms.GA.Cluster): the next cluster in the collections.
		"""
		while True:
			if self.index_collections < len(self.collections):
				if self.index < len(self.collections[self.index_collections]):
					cluster = self.collections[self.index_collections][self.index]
					self.index += 1
					return cluster
				elif self.index == len(self.collections[self.index_collections]):
					self.index = 0
					self.index_collections += 1
				else:
					print('error')
					exit()
			elif self.index_collections == len(self.collections):
				raise StopIteration
			else:
				print('error')
				exit()



import sys
if sys.version_info[0] == 2:
	raise ImportError('The Organisms program requires Python3. This is Python2.')

__name__    = 'The Otago Research Genetic Algorithm for Nanoclusters, Including Structural Methods and Similarity (Organisms) Program'
__version__ = '3.2.4'
__author__  = 'Geoffrey Weal and Dr. Anna Garden'

__author_email__ = 'anna.garden@otago.ac.nz'
__license__ = 'GNU AFFERO GENERAL PUBLIC LICENSE'
__url__ = 'https://github.com/GardenGroupUO/Organisms'
__doc__ = 'See https://organisms.readthedocs.io/en/latest/ for the documentation on this program'

from Organisms.GA.GA_Program import GA_Program
from Organisms.Subsidiary_Programs.MakeTrialsProgram import MakeTrialsProgram
#from Organisms.Subsidiary_Programs.GetNewlyInitilisedPopulation import GetNewlyInitilisedPopulation
__all__ = ['GA_Program','MakeTrialsProgram'] # 'GetNewlyInitilisedPopulation']



from Organisms.GA.SCM_Scripts.SCM_initialisation import get_rCuts, get_SCM_methods
from Organisms.GA.SCM_Scripts.CNA_Database import CNA_Database

from Organisms.GA.Fitness_Operators.CNA_Database_Check import CNA_Database_Check
from Organisms.GA.Fitness_Operators.Fitness_Function import Fitness_Function

from Organisms.GA.Fitness_Operators.CNA_Fitness_Contribution import get_CNA_fitness_contribution, get_CNA_fitness_contribution_normalised, mean

def SCM_options(self,fitness_information, predation_operator, no_of_cpus):
	"""
	This method records the information from the predation_operator that is relavant to the fitness class rather than the needed to obtain the predation operator class, and make a link to the similarity profile database, containing the similarity plots of every cluster against every other cluster for the current generation,

	:param fitness_information: This is all the information that is needed about the fitness class
	:type  fitness_information: dict.
	:param predation_operator: This is the predation operator that this fitness class will take information from if needed to obtain a fitness.
	:type  predation_operator: Organisms.GA.Predation_Operator
	:param no_of_cpus: the number of cpus to use when multiprocessing
	:type  no_of_cpus: int

	"""
	self.predation_switch = predation_operator.Predation_Switch
	if not ('SCM Scheme' in fitness_information) and not (fitness_information['SCM Scheme'] in ['T-SCM','A-SCM']):
		print('ERROR in def SCM_options, in Class SCM_and_Energy_Fitness_Operator, in SCM_and_Energy_Fitness_Operator.py/SCM_and_Energy_Fitness_Operator_Options.py')
		print('You either need to include the "SCM Scheme" setting, or you have entered an invalid entry for "SCM Scheme" setting in the fitness_information dictionary.')
		print('The "SCM Scheme" setting can be either the T-SCM or the A-SCM.')
		print('Your fitness_information dictionary is: '+str(fitness_information))
		print('Check this.')
		exit('This program will finish without completing.')
	# --------------------------------------------------- #
	# Settings for the 'Use Predation Information' setting
	# The end result of this is to set up the cna_database.
	will_make_new_cna_database = CNA_Database_Check(self,fitness_information, predation_operator)
	if will_make_new_cna_database:
		self.cna_database = make_new_cna_database(self, fitness_information, no_of_cpus)
	# --------------------------------------------------- #
	# Get the other variable for the SCM + energy fitness operator from the fitness_information dictionary.
	self.fitness_switch = fitness_information['Fitness Operator']
	self.SCM_fitness_contribution = fitness_information['SCM_fitness_contribution']
	self.dynamic_mode = fitness_information['Dynamic Mode']
	# --------------------------------------------------- #
	# Get the fitness contributions
	self.energy_fitness_contribution = 1.0 - self.SCM_fitness_contribution
	self.fitness_weights = []
	if not self.energy_fitness_contribution == 0.0:
		self.fitness_weights.append(self.energy_fitness_contribution)
	if not self.SCM_fitness_contribution == 0.0:
		self.fitness_weights.append(self.SCM_fitness_contribution)
	# Double check the fitness_contributions add up to 1
	if not sum(self.fitness_weights) == 1.0:
		print('Error in def get_fitness of class SCM_Predation_Operator, in SCM_Predation_Operator.py')
		print('The sum of the coefficients of the linear superposition (sorry cant remember the right word right now) does not equal 1.')
		print('Coefficients: '+str(self.fitness_weights))
		print('sum(fitness_weights) = '+str(sum(self.fitness_weights)))
		print('Check this.')
		import pdb; pdb.set_trace()
		exit('This program will finish without completing')
	# --------------------------------------------------- #
	# Determine how the collection of sigma values between cluster x and every other cluster in the population (including offspring.)
	if 'Take from the collection of a clusters similarities' in fitness_information:
		self.take_for_clusters_similarities = fitness_information['Take from the collection of a clusters similarities']
	else:
		self.take_for_clusters_similarities = 'Maximum'
	if self.take_for_clusters_similarities == 'Maximum':
		self.collection_function = max
	elif self.take_for_clusters_similarities == 'Average':
		self.collection_function = mean
	else:
		print('Error in def get_fitness of class SCM_Predation_Operator, in SCM_Predation_Operator.py')
		print('You have not supplied a valid input for "Take from the collection of a clusters similarities"')
		print('"Take from the collection of a clusters similarities" can only be either "Maximum" or "Average"')
		print('You input for "Take from the collection of a clusters similarities": '+str(self.take_for_clusters_similarities))
		print('Check this.')
		exit('This program will finish without completing')
	# --------------------------------------------------- #
	# Determine if the similarities used to obtain the structure + energy fitness operator are normalised or not
	if 'normalise_similarities' in fitness_information:
		self.normalise_similarities = fitness_information['normalise_similarities']
	else:
		self.normalise_similarities = False
	if self.normalise_similarities:
		self.get_CNA_fitness_contribution = get_CNA_fitness_contribution_normalised
	else:
		self.get_CNA_fitness_contribution = get_CNA_fitness_contribution
	# --------------------------------------------------- #
	# fitness functions
	self.energy_fitness_function = Fitness_Function(**fitness_information['energy_fitness_function'])
	self.cna_fitness_function    = Fitness_Function(**fitness_information['SCM_fitness_function'])
	if self.dynamic_mode:
		self.population_fitness_function = Fitness_Function(**fitness_information['fitness_function'])
	# --------------------------------------------------- #

def make_new_cna_database(self, fitness_information, no_of_cpus):
	"""
	Reset the CNA database

	:param fitness_information: This is all the information that is needed about the fitness class
	:type  fitness_information: dict.
	:param no_of_cpus: the number of cpus to use when multiprocessing
	:type  no_of_cpus: int

	:returns: return a CNA_Database object
	:rtype:   Organisms.GA.SCM_Scripts.CNA_Database
	"""
	SCM_Scheme = fitness_information['SCM Scheme']
	rCuts = get_rCuts(self,fitness_information)
	get_cna_profile_method, get_cna_similarities_method = get_SCM_methods(SCM_Scheme)
	try:
		cut_off_similarity = predation_Information['Cut_off']
	except:
		cut_off_similarity = 100.0
	debug = False
	return CNA_Database(rCuts,self.population,cut_off_similarity,get_cna_profile_method,get_cna_similarities_method,no_of_cpus,debug)



import os, time

class GA_Program_Details:
	"""
	This class is designed to keep track of the process of the genetic algorithm run.

	:param GA_Program: This is the main genetic algorithm program this class will be writing about
	:type  GA_Program: Organisms.GA.GA_Program: 
	:param is_new_ga: This determines if to write all the details about this genetic algorithm to a file. You only need to do this if the genetic algorithm is just beginning from generation 0. Default: True.
	:type  is_new_ga: bool.
	"""
	def __init__(self,GA_Program,is_new_ga=True):
		self.Run_path = GA_Program.Run_path
		self.details_filename = 'GA_Run_Details.txt'
		# Recording data about this genetic algorithm run.
		self.number_of_matings = 0
		self.number_of_successful_matings = 0
		#self.number_of_mutation = [0]*len(GA_Program.mutTypes)
		#self.number_of_successful_mutation = [0]*len(GA_Program.mutTypes)
		self.no_of_explosions = 0
		self.no_of_not_converged = 0
		self.create(is_new_ga)

	##############################################################################################################

	def create(self, is_new_ga): #add new variable
		"""
		This definition will create a file containing the details of this genetic algorithm on the disk.

		:param is_new_ga: This determines if to write all the details about this genetic algorithm to a file. You only need to do this if the genetic algorithm is just beginning from generation 0.
		:type  is_new_ga: bool.
		"""
		if is_new_ga:
			if not os.path.exists(self.Run_path+'/'+self.details_filename):
				self.open('w')
				self.close()
			else:
				print('Error in def create of class GAProgram_Details, in GAProgram_Details.py')
				print(str(self.details_filename)+' in '+str(self.Run_path)+' already exists.')
				print('Check this.')
				print('The program is finishing without performing any part of the genetic algorithm.')
				exit() 
		else:
			if not os.path.exists(self.Run_path+'/'+self.details_filename):
				print('Error in def create of class GAProgram_Details, in GAProgram_Details.py')
				print(str(self.details_filename)+' in '+str(self.Run_path)+' does not exist, however the GA is being resumed, so should already have a '+str(self.details_filename)+' file.')
				print('Check this.')
				print('The program is finishing without performing any part of the genetic algorithm.')
				exit() 

	def start_clock(self):
		"""
		Get the Starting time for this genetic algorithm, which will be used to time each generation.
		"""
		self.startTime = time.time()

	def end_clock(self, generation):
		"""
		Get the end time taken for a generation to run.

		:param generation: The generation that was run. 
		:type  generation: int

		returns time_taken: The time taken for the generation to run in seconds. 
		rtype   time_taken: float
		"""
		endTime = time.time()
		time_taken = endTime - self.startTime
		del self.startTime
		self.open('a')
		self.GA_Run_Details.write(str(generation)+': '+str(time_taken)+' s\n')
		self.close()
		return time_taken

	def ending_details(self):
		"""
		This def will write the finishing remarks of the genetic algorithm to GA_Run_Details
		"""
		self.endTime = time.time()

		self.open("w+")
		self.GA_Run_Details.write("The Genetic Algorithm finished successfully.\n")
		self.GA_Run_Details.write("This program took " + str((self.endTime - self.startTime)) + " s to run.\n")
		self.GA_Run_Details.write("\n")
		self.GA_Run_Details.write("DETAILS\n")
		self.GA_Run_Details.write("\n")
		self.GA_Run_Details.write("Number of Instances where cluster creation was restarted due to the cluster exploding: "+str(self.no_of_explosions)+"\n")
		self.GA_Run_Details.write("Number of Instances where cluster creation was restarted due to the local optimisation not converging: "+str(self.no_of_not_converged)+"\n")
		self.close()

	##############################################################################################################
		
	def open(self,read_type):
		"""
		This will open the GA_Run_Details text file that is being recorded to.

		This is designed to be a private definition

		:param read_type: This is the way to open the file, either read (r), write (w), or append (a).
		:type  read_type: char
		"""
		self.GA_Run_Details = open(self.Run_path+'/'+self.details_filename, str(read_type))

	def close(self):
		"""
		This will close the GA_Run_Details text file that is being recorded to.

		This is designed to be a private definition
		"""
		self.GA_Run_Details.close()



#!/usr/bin/python

EnergyProfileTXT_path = 'Population/Population_history.txt'
recorded_current_population_no = 1
all_energies = []
restart_generations = []
restart_generations_plot_marker = []
no_of_restarts = 0
with open(EnergyProfileTXT_path,'r') as EnergyProfileTXT:
	for NOUSED in range(5):
		EnergyProfileTXT.readline()
	counter = 0
	for line in EnergyProfileTXT:
		if '#-----------------------------------------------------------------------------' in line:
			continue
		elif 'Reset population as reached epoch' in line:
			print('Population Restarted')
			restart_generations.append(current_generation_no)
			restart_generations_plot_marker.append(current_generation_no+no_of_restarts+0.5)
			no_of_restarts += 1
			continue
		if counter == 0:
			current_generation_no = int(line.replace('GA Iteration:',''))
			print('Generation: '+str(current_generation_no))
			if current_generation_no == recorded_current_population_no:
				exit('Error')
			recorded_current_population_no += 1
		elif counter == 1:
			line.replace('Clusters in Pool:\t','').split()
		elif counter == 2:
			energies = tuple(float(energy) for energy in line.replace('Energies of Clusters:\t','').split())
			all_energies.append(energies)
			counter = -1
		counter += 1

plot_names = []
plot_energies = []
for index in range(len(all_energies)):
	energies = all_energies[index]
	plot_names += [index+1]*len(energies)
	plot_energies += energies

# Make plot.
import matplotlib.pyplot as plt
plt.scatter(plot_names, plot_energies, s=2, c='b', alpha=0.5)
plt.xlabel('Generation')
plt.ylabel('Energy')
plt.tight_layout()
plt.savefig('energy_of_population_over_generation.png')